from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
"""Classes for evaluating germ set performance."""

import numpy as _np
import matplotlib.pyplot as plt

class GermSetEval:
    def __init__(self, germset=None, gatesets=None, resultDict=None,
                 errorDict=None):
        self.germset = germset
        self.gatesets = gatesets
        self.resultDict = resultDict
        self.errorDict = errorDict

    def calc_marg_pows(self):
        if self.errorDict is None:
            raise ValueError('Dictionary of estimate errors not present, so '
                             'cannot calculate marginal powers!')
        marginalPowers = {}
        for (gsNum, numClicks, run), convergence in self.errorDict.items():
            for (error0, L0), (error1, L1) in zip(convergence[:-1],
                                                  convergence[1:]):
                if (numClicks, L1) in marginalPowers:
                    marginalPowers[numClicks, L1].append(_np.log(error1/error0)
                                                         / _np.log(L1/L0))
                else:
                    marginalPowers[numClicks, L1] = [_np.log(error1/error0)
                                                     / _np.log(L1/L0)]
        return marginalPowers

    def plot_marg_pows(self, axs=None):
        if self.errorDict is None:
            raise ValueError('Dictionary of estimate errors not present, so '
                             'cannot plot marginal powers!')

        clickNums = set()
        for trueGatesetNum, numClicks, run in self.errorDict:
            clickNums.add(numClicks)
        clickNums = sorted(list(clickNums))
        numClickNums = len(clickNums)
        if axs is None:
            fig = plt.figure(figsize=(6, 4*numClickNums))
            axs = []
            for row in range(1, numClickNums + 1):
                axs.append(fig.add_subplot(numClickNums, 1, row))
        elif len(axs) != numClickNums:
            raise ValueError("The number of axs provided must be equal to the "
                             "number of unique click numbers ({})!"
                             .format(numClickNums))

        marginalPowers = self.calc_marg_pows()
        for row, (ax, currentNumClicks) in enumerate(zip(axs, clickNums)):
            Ls = sorted([L for numClicks, L in marginalPowers.keys()
                         if numClicks == currentNumClicks])
            ax.boxplot([marginalPowers[currentNumClicks, L] for L in Ls],
                       whis=_np.inf)
            ax.set_ylabel(r'Marginal power ($\alpha$)', fontsize=18)
            ax.set_xticklabels(Ls)
            if row == numClickNums - 1:
                ax.set_xlabel('Sequence length ($L$)', fontsize=18)
            ax.set_title(r'{} clicks'.format(currentNumClicks), fontsize=18)

        return axs

    def plot_spectrum(self, spectrum, numGaugeParams=None, ax=None):
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
            return_fig_ax = True
        else:
            return_fig_ax = False

        if numGaugeParams is None:
            legend = False
            numGaugeParams = 0
        else:
            legend = True

        ax.semilogy(range(numGaugeParams + 1, len(spectrum) + 1),
                    spectrum[numGaugeParams:], 'o',
                    label='Physical parameters')
        ax.semilogy(range(1, numGaugeParams + 1), spectrum[:numGaugeParams],
                    'o', label='Gauge parameters', color='r')
        if legend:
            ax.legend(loc=4, framealpha=0.5)
        if return_fig_ax:
            return {'fig': fig, 'ax': ax}

    def plot_spectra(self, axs=None):
        from ..algorithms import germselection as germsel
        
        missing = [key for key in self.__dict__
                   if key in ['germset', 'gatesets']
                   and self.__dict__[key] is None]
        if len(missing) > 0:
            raise ValueError('Missing {}, so cannot plot spectra!'
                             .format(' and '.join(missing)))

        numGatesets = len(self.gatesets)
        numGaugeParams = germsel.removeSPAMVectors(
            self.gatesets[0]).num_gauge_params()
        if axs is None:
            fig = plt.figure(figsize=(6, 4*numGatesets))
            axs = []
            for row in range(1, numGatesets + 1):
                axs.append(fig.add_subplot(numGatesets, 1, row))
        elif len(axs) != numGatesets:
            raise ValueError("The number of axs provided must be equal to the "
                             "number of gatesets ({})!".format(numGatesets))

        for gatesetNum, (ax, gateset) in enumerate(zip(axs, self.gatesets)):
            spectrum = germsel.test_germ_list_infl(gateset, self.germset,
                                                   scoreFunc='all',
                                                   returnSpectrum=True)[1]
            self.plot_spectrum(spectrum, numGaugeParams, ax)
            ax.set_title('GateSet {}'.format(gatesetNum))

        return axs


    def calc_ranges(self):
        if self.errorDict is None:
            raise ValueError('Dictionary of estimate errors not present, so '
                             'cannot calculate error ranges!')
        errorsAtL = {}
        for (gsNum, numClicks, run), convergence in self.errorDict.items():
            for error, L in convergence:
                if (numClicks, L) in errorsAtL:
                    errorsAtL[numClicks, L].append(error)
                else:
                    errorsAtL[numClicks, L] = [error]

        ranges = {key: _np.percentile(_np.array(value), [0.0, 100.0])
                  for key, value in errorsAtL.items()}

        return ranges


    def plot_ranges(self, axs=None):
        if self.errorDict is None:
            raise ValueError('Dictionary of estimate errors not present, so '
                             'cannot plot error ranges!')

        clickNums = set()
        for trueGatesetNum, numClicks, run in self.errorDict:
            clickNums.add(numClicks)
        clickNums = sorted(list(clickNums))
        numClickNums = len(clickNums)
        if axs is None:
            fig = plt.figure(figsize=(6, 4*numClickNums))
            axs = []
            for row in range(1, numClickNums + 1):
                axs.append(fig.add_subplot(numClickNums, 1, row))
        elif len(axs) != numClickNums:
            raise ValueError("The number of axs provided must be equal to the "
                             "number of unique click numbers ({})!"
                             .format(numClickNums))

        ranges = self.calc_ranges()
        for row, (ax, currentNumClicks) in enumerate(zip(axs, clickNums)):
            Ls = sorted([L for numClicks, L in ranges.keys()
                         if numClicks == currentNumClicks])
            min_dists, max_dists = zip(*[ranges[currentNumClicks, L]
                                         for L in Ls])
            ax.fill_between(Ls, min_dists, max_dists, alpha=0.5,
                            label='{} clicks'.format(int(currentNumClicks)),
                            linewidth=0)
            ax.set_yscale('log')
            ax.set_xscale('log')
            ax.set_ylabel(r'Distance to truth', fontsize=18)
            ax.set_xticks(Ls)
            ax.set_xticklabels(Ls)
            if row == numClickNums - 1:
                ax.set_xlabel('Sequence length ($L$)', fontsize=18)
            ax.set_title(r'{} clicks'.format(currentNumClicks), fontsize=18)

        return axs



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
"""Functions to facilitate using GRASP."""

import itertools
import random

import numpy as _np

from .. import objects as _objs


def get_swap_neighbors(weights, forcedWeights=None, shuffle=False):
    """Return the list of weights in the neighborhood of a given weight vector.

    A weight vector is in the neighborhood of a given weight vector if it is
    only a single swap away from the given weight vector. There is an option to
    use `forcedWeights` to indicate elements you don't want to swap out.

    Parameters
    ----------
    weights : numpy.array
        Binary vector to find the neighborhood of.

    forcedWeights : numpy.array, optional
        Binary vector indicating elements that must be included in all
        neighboring vectors (these elements are assumed to already be present
        in `weights`.

    shuffle : bool, optional
        Whether the neighborhood should be presented to the optimizer in a
        random order (important if the local optimizer updates the solution to
        the first better solution it finds in the neighborhood instead of
        exhaustively searching the neighborhood for the best solution).

    Returns
    -------
    list of numpy.array
        List of binary vectors corresponding to all the neighbors of `weights`.

    """
    if forcedWeights is None:
        forcedWeights = _np.zeros(len(weights))

    swap_out_idxs = _np.where(_np.logical_and(weights == 1,
                                              forcedWeights == 0))[0]
    swap_in_idxs = _np.where(weights == 0)[0]
    neighbors = []
    for swap_out, swap_in in itertools.product(swap_out_idxs, swap_in_idxs):
        neighbor = weights.copy()
        neighbor[swap_out] = 0
        neighbor[swap_in] = 1
        neighbors.append(neighbor)

    if shuffle:
        random.shuffle(neighbors)

    return neighbors


def grasp_greedy_construction(elements, scoreFn, rclFn, feasibleThreshold=None,
                              feasibleFn=None, initialElements=None,
                              seed=None):
    if initialElements is None:
        weights = _np.zeros(len(elements))
    else:
        if len(initialElements) != len(elements):
            raise ValueError('initialElements must have the same length as '
                             'elements ({}), not {}!'.format(len(elements),
                                                        len(initialElements)))
        weights = _np.array(initialElements)

    soln = [elements[idx] for idx in _np.nonzero(weights)[0]]

    if feasibleThreshold is not None:
        feasibleTest = 'threshold'
    elif feasibleFn is not None:
        feasibleTest = 'function'
    else:
        raise ValueError('Must provide either feasibleFn or '
                         'feasibleThreshold!')

    feasible = False

    while _np.any(weights==0) and not feasible:
        candidateIdxs = _np.where(weights==0)[0]
        candidateSolns = [soln + [elements[idx]] for idx in candidateIdxs]
        candidateScores = _np.array([scoreFn(candidateSoln)
                                     for candidateSoln in candidateSolns])
        rclIdxs = rclFn(candidateScores)
        chosenIdx = _np.random.choice(rclIdxs)
        soln = candidateSolns[chosenIdx]
        weights[candidateIdxs[chosenIdx]] = 1
        if feasibleTest == 'threshold':
            feasible = candidateScores[chosenIdx] <= feasibleThreshold
        elif feasibleTest == 'function':
            feasible = feasibleFn(soln)

    if not feasible:
        raise ValueError('No feasible solution found!')

    return soln


def grasp_local_search(initialSoln, scoreFn, elements, getNeighborsFn,
                       feasibleThreshold=None, feasibleFn=None):

    if feasibleThreshold is not None:
        feasibleTest = 'threshold'
    elif feasibleFn is not None:
        feasibleTest = 'function'
    else:
        raise ValueError('Must provide either feasibleFn or '
                         'feasibleThreshold!')

    currentSoln = initialSoln
    currentWeights = _np.zeros(len(elements))
    for element in initialSoln:
        currentWeights[elements.index(element)] = 1
    currentScore = scoreFn(currentSoln)

    betterSolnFound = True

    while betterSolnFound:
        betterSolnFound = False
        weightsNeighbors = getNeighborsFn(currentWeights)
        neighborSolns = [[element for element
                          in _np.array(elements)[_np.nonzero(weightsNeighbor)]]
                         for weightsNeighbor in weightsNeighbors]
        if feasibleTest == 'function':
            feasibleNeighborSolns = [(idx, soln) for idx, soln
                                     in enumerate(neighborSolns)
                                     if feasibleFn(soln)]
            for idx, soln in feasibleNeighborSolns:
                solnScore = scoreFn(soln)
                if solnScore < currentScore:
                    betterSolnFound = True
                    currentScore = solnScore
                    currentSoln = soln
                    currentWeights = weightsNeighbors[idx]
                    break

        elif feasibleTest == 'threshold':
            for idx, soln in enumerate(neighborSolns):
                solnScore = scoreFn(soln)
                # The current score is by construction below the threshold,
                # so we don't need to check that.
                if solnScore < currentScore:
                    betterSolnFound = True
                    currentScore = solnScore
                    currentSoln = soln
                    currentWeights = weightsNeighbors[idx]
                    break

    return currentSoln


def do_grasp_iteration(elements, greedyScoreFn, rclFn, localScoreFn,
                       getNeighborsFn, feasibleThreshold=None, feasibleFn=None,
                       initialElements=None, seed=None, verbosity=0):
    """Perform one iteration of GRASP (greedy construction and local search).

    Parameters
    ----------
    elements : list
        A list containing some representation of the elements that can be used
        by the verious score functions.

    greedyScoreFn : callable
        Function that takes a sublist of `elements` and returns a score to
        minimize that is comparable with other scores. Used by the greedy
        construction to construct the initial feasible subset.

    rclFn : callable
        Function that takes a list of sublists of `elements` (that is, a list
        of candidate partial solutions) and returns the indices within that
        list of partial solutions to be included in the restricted candidate
        list.

    localScoreFn : callable
        Function that takes a sublist of `elements` and returns a score to
        minimize that is comparable with other scores. Used by the local
        search to find a locally optimal feasible subset.

    getNeighborsFn : callable
        Function that takes a binary vector indicating which members of
        `elements` are included in the current solution and returns a list
        of binary vectors indicating which potential solutions are in the
        neighborhood of the current solution for the purposes of local
        optimization.

    feasibleThreshold : score
        A value comparable with the return value of the various score functions
        against which a score may be compared to test if the solution is
        feasible (the solution is feasible iff
        ``solnScore < feasibleThreshold``). Overrides `feasibleFn` if set to a
        value other than ``None``.

    feasibleFn : callable
        Function that takes a sublist of `elements` defining a potential
        solution and returns ``True`` if that solution is feasible (otherwise
        should return ``False``). Not used if `feasibleThreshold` is not
        ``None``.

    initialElements : numpy.array
        Binary vector indicating whether the corresponding elements in
        `elements` should be automatically included by the greedy construction
        routine at the start of its construction.

    seed : int
        Seed for the random number generator.

    verbosity : int
        Sets the level of logging messages the printer will display.

    Returns
    -------
    initialSoln : list
        The sublist of `elements` given by the greedy construction.

    localSoln : list
        The sublist of `elements` given by the local search.

    """
    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    initialSoln = grasp_greedy_construction(elements, greedyScoreFn, rclFn,
                                            feasibleThreshold, feasibleFn,
                                            initialElements, seed)
    printer.log('Initial construction:', 1)
    printer.log(str([str(element) for element in initialSoln]), 1)

    localSoln = grasp_local_search(initialSoln, localScoreFn, elements,
                                   getNeighborsFn, feasibleThreshold,
                                   feasibleFn)
    printer.log('Local optimum:', 1)
    printer.log(str([str(element) for element in localSoln]), 1)

    return initialSoln, localSoln


def grasp(elements, greedyScoreFn, rclFn, localScoreFn, getNeighborsFn,
          finalScoreFn, iterations, feasibleThreshold=None, feasibleFn=None,
          initialElements=None, seed=None, verbosity=0):
    """Perform GRASP to come up with an optimal feasible set of elements.

    Parameters
    ----------
    elements : list
        A list containing some representation of the elements that can be used
        by the verious score functions.

    greedyScoreFn : callable
        Function that takes a sublist of `elements` and returns a score to
        minimize that is comparable with other scores. Used by the greedy
        construction to construct the initial feasible subset.

    rclFn : callable
        Function that takes a list of sublists of `elements` (that is, a list
        of candidate partial solutions) and returns the indices within that
        list of partial solutions to be included in the restricted candidate
        list.

    localScoreFn : callable
        Function that takes a sublist of `elements` and returns a score to
        minimize that is comparable with other scores. Used by the local
        search to find a locally optimal feasible subset.

    getNeighborsFn : callable
        Function that takes a binary vector indicating which members of
        `elements` are included in the current solution and returns a list
        of binary vectors indicating which potential solutions are in the
        neighborhood of the current solution for the purposes of local
        optimization.

    finalScoreFn : callable
        Function that takes a sublist of `elements` and returns a score to
        minimize that is comparable with other scores. Used to compare the
        solutions from various iterations in order to choose an optimum.

    iterations : int
        How many iterations of greedy construction followed by local search to
        perform.

    feasibleThreshold : score
        A value comparable with the return value of the various score functions
        against which a score may be compared to test if the solution is
        feasible (the solution is feasible iff
        ``solnScore < feasibleThreshold``). Overrides `feasibleFn` if set to a
        value other than ``None``.

    feasibleFn : callable
        Function that takes a sublist of `elements` defining a potential
        solution and returns ``True`` if that solution is feasible (otherwise
        should return ``False``). Not used if `feasibleThreshold` is not
        ``None``.

    initialElements : numpy.array
        Binary vector with 1s at indices corresponding to elements in
        `elements` that the greedy construction routine will include at the
        start of its construction.

    seed : int
        Seed for the random number generator.

    verbosity : int
        Sets the level of logging messages the printer will display.

    Returns
    -------
    list of GateString
        The best germ set from all locally-optimal germ sets constructed.

    """
    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    bestSoln = None
    for iteration in range(iterations):
        printer.log('Iteration {}'.format(iteration), 1)
        initialSoln, localSoln = do_grasp_iteration(elements, greedyScoreFn,
                                                    rclFn, localScoreFn,
                                                    getNeighborsFn,
                                                    feasibleThreshold,
                                                    feasibleFn,
                                                    initialElements, seed,
                                                    verbosity)
        if bestSoln is None:
            bestSoln = localSoln
        elif finalScoreFn(localSoln) < finalScoreFn(bestSoln):
            bestSoln = localSoln

    return bestSoln



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
"""Functions for the construction of new gate sets."""

import numpy as _np
import itertools as _itertools
import collections as _collections
import scipy.linalg as _spl

from ..tools import basistools as _bt
from ..objects import gate as _gate
from ..objects import gateset as _gateset
from ..objects import gaugegroup as _gg


#############################################
# Build gates based on "standard" gate names
############################################

#TODO: stateSpaceLabels is never used?
def build_vector(stateSpaceDims, stateSpaceLabels, vecExpr, basis="gm"):
    """
    Build a rho or E vector from an expression.

    Parameters
    ----------
    stateSpaceDims : list of ints
        Dimensions specifying the structure of the density-matrix space.
        Elements correspond to block dimensions of an allowed density matrix in
        the standard basis, and the density-matrix space is the direct sum of
        linear spaces of dimension block-dimension^2.

    stateSpaceLabels : a list of tuples
        Each tuple corresponds to a block of a density matrix in the standard
        basis (and therefore a component of the direct-sum density matrix
        space). Elements of a tuple are user-defined labels beginning with "L"
        (single level) or "Q" (two-level; qubit) which interpret the
        d-dimensional state space corresponding to a d x d block as a tensor
        product between qubit and single level systems.

    vecExpr : string
        the expression which determines which vector to build.  Currenlty, only
        integers are allowed, which specify a the vector for the pure state of
        that index.  For example, "1" means return vectorize(``|1><1|``).  The
        index labels the absolute index of the state within the entire state
        space, and is independent of the direct-sum decomposition of density
        matrix space.

    basis : {'gm','pp','std','qt'}, optional
       the basis of the returned vector.

        - 'std' == Standard (matrix units)
        - 'gm' == Gell-Mann
        - 'pp' == Pauli-product
        - 'qt' == Qutrit

    Returns
    -------
    numpy array
        The vector specified by vecExpr in the desired basis.
    """
    _, gateDim, blockDims = _bt._processBlockDims(stateSpaceDims)
    vecInReducedStdBasis = _np.zeros( (gateDim,1), 'd' ) # assume index given as vecExpr refers to a
                                                         #Hilbert-space state index, so "reduced-std" basis

    #So far just allow integer prepExpressions that give the index of state (within the state space) that we prep/measure
    try:
        index = int(vecExpr)
    except:
        raise ValueError("Expression must be the index of a state (as a string)")

    start = 0; vecIndex = 0
    for blockDim in blockDims:
        for i in range(start,start+blockDim):
            for j in range(start,start+blockDim):
                if (i,j) == (index,index):
                    vecInReducedStdBasis[ vecIndex, 0 ] = 1.0  #set diagonal element of density matrix
                    break
                vecIndex += 1
        start += blockDim

    return _bt.change_basis(vecInReducedStdBasis, "std", basis, stateSpaceDims)

def build_identity_vec(stateSpaceDims, basis="gm"):
    """
    Build a the identity vector for a given space and basis.

    Parameters
    ----------
    stateSpaceDims : list of ints
       Dimenstions specifying the structure of the density-matrix space.
        Elements correspond to block dimensions of an allowed density matrix in
        the standard basis, and the density-matrix space is the direct sum of
        linear spaces of dimension block-dimension^2.

    basis : {'gm','pp','std','qt'}, optional
        the basis of the returned vector.

        - 'std' == Standard (matrix units)
        - 'gm' == Gell-Mann
        - 'pp' == Pauli-product
        - 'qt' == Qutrit

    Returns
    -------
    numpy array
        The identity vector in the desired basis.
    """
    _, gateDim, blockDims = _bt._processBlockDims(stateSpaceDims)
    vecInReducedStdBasis = _np.zeros( (gateDim,1), 'd' ) # assume index given as vecExpr refers to a Hilbert-space state index, so "reduced-std" basis

    #set all diagonal elements of density matrix to 1.0 (end result = identity density mx)
    start = 0; vecIndex = 0
    for blockDim in blockDims:
        for i in range(start,start+blockDim):
            for j in range(start,start+blockDim):
                if i == j: vecInReducedStdBasis[ vecIndex, 0 ] = 1.0  #set diagonal element of density matrix
                vecIndex += 1
        start += blockDim

    return _bt.change_basis(vecInReducedStdBasis, "std", basis, stateSpaceDims)



def _oldBuildGate(stateSpaceDims, stateSpaceLabels, gateExpr, basis="gm"):
#coherentStateSpaceBlockDims
    """
    Build a gate matrix from an expression

    Parameters
    ----------
    stateSpaceDims : a list of integers specifying the dimension of each block
    of a block-diagonal the density matrix
    stateSpaceLabels : a list of tuples, each one corresponding to a block of
    the density matrix.  Elements of the tuple are user-defined labels
    beginning with "L" (single level) or "Q" (two-level; qubit) which interpret
    the states within the block as a tensor product structure between the
    labelled constituent systems.

    gateExpr : string containing an expression for the gate to build

    basis : string
      - "std" = gate matrix operates on density mx expressed as sum of matrix
        units
      - "gm"  = gate matrix operates on dentity mx expressed as sum of
        normalized Gell-Mann matrices
      - "pp"  = gate matrix operates on density mx expresses as sum of
        tensor-prod of pauli matrices
    """
    # gateExpr can contain single qubit ops: X(theta) ,Y(theta) ,Z(theta)
    #                      two qubit ops: CNOT
    #                      clevel qubit ops: Leak
    #                      two clevel opts: Flip
    #  each of which is given additional parameters specifying which indices it acts upon


    #Gate matrix will be in matrix unit basis, which we order by vectorizing
    # (by concatenating rows) each block of coherent states in the order given.
    dmDim, _ , _ = _bt._processBlockDims(stateSpaceDims)
    fullOpDim = dmDim**2

    #Store each tensor product blocks start index (within the density matrix), which tensor product block
    #  each label is in, and check to make sure dimensions match stateSpaceDims
    tensorBlkIndices = {}; startIndex = []; M = 0
    assert( len(stateSpaceDims) == len(stateSpaceLabels) )
    for k, blockDim in enumerate(stateSpaceDims):
        startIndex.append(M); M += blockDim

        #Make sure tensor-product interpretation agrees with given dimension
        tensorBlkDim = 1 #dimension of this coherent block of the *density matrix*
        for s in stateSpaceLabels[k]:
            tensorBlkIndices[s] = k
            if s.startswith('Q'): tensorBlkDim *= 2
            elif s.startswith('L'): tensorBlkDim *= 1
            else: raise ValueError("Invalid state space specifier: %s" % s)
        if tensorBlkDim != blockDim:
            raise ValueError("State labels %s for tensor product block %d have dimension %d != given dimension %d" \
                                 % (stateSpaceLabels[k], k, tensorBlkDim, blockDim))


    #print "DB: dim = ",dim, " dmDim = ",dmDim
    gateInStdBasis = _np.identity( fullOpDim, 'complex' )
      # in full basis of matrix units, which we later reduce to the
      # that basis of matrix units corresponding to the allowed non-zero
      #  elements of the density matrix.

    exprTerms = gateExpr.split(':')
    for exprTerm in exprTerms:

        gateTermInStdBasis = _np.identity( fullOpDim, 'complex' )
        l = exprTerm.index('('); r = exprTerm.index(')')
        gateName = exprTerm[0:l]
        argsStr = exprTerm[l+1:r]
        args = argsStr.split(',')

        if gateName == "I":
            pass

        elif gateName in ('X','Y','Z'): #single-qubit gate names
            assert(len(args) == 2) # theta, qubit-index
            theta = eval( args[0], {"__builtins__":None}, {'pi': _np.pi})
            label = args[1].strip(); assert(label.startswith('Q'))

            if gateName == 'X': ex = -1j * theta*_bt.sigmax/2
            elif gateName == 'Y': ex = -1j * theta*_bt.sigmay/2
            elif gateName == 'Z': ex = -1j * theta*_bt.sigmaz/2
            Ugate = _spl.expm(ex) # 2x2 unitary matrix operating on single qubit in [0,1] basis

            iTensorProdBlk = tensorBlkIndices[label] # index of tensor product block (of state space) this bit label is part of
            cohBlk = stateSpaceLabels[iTensorProdBlk]
            basisInds = []
            for l in cohBlk:
                assert(l[0] in ('L','Q')) #should have been checked above
                if l.startswith('L'): basisInds.append([0])
                elif l.startswith('Q'): basisInds.append([0,1])

            tensorBlkBasis = list(_itertools.product(*basisInds))
            K = cohBlk.index(label)
            N = len(tensorBlkBasis)
            UcohBlk = _np.identity( N, 'complex' ) # unitary matrix operating on relevant tensor product block part of state
            for i,b1 in enumerate(tensorBlkBasis):
                for j,b2 in enumerate(tensorBlkBasis):
                    if (b1[:K]+b1[K+1:]) == (b2[:K]+b2[K+1:]):   #if all part of tensor prod match except for qubit we're operating on
                        UcohBlk[i,j] = Ugate[ b1[K], b2[K] ] # then fill in element

            UcohBlkc = UcohBlk.conjugate()
            gateBlk = _np.kron(UcohBlk,UcohBlkc) # N^2 x N^2 mx operating on vectorized tensor product block of densty matrix

            #Map gateBlk's basis into final gate basis
            mapBlk = []
            s = startIndex[iTensorProdBlk] #within state space (i.e. row or col of density matrix)
            cohBlkSize = UcohBlk.shape[0]
            for i in range(cohBlkSize):
                for j in range(cohBlkSize):
                    vec_ij_index = (s+i)*dmDim + (s+j) #vectorize by concatenating rows
                    mapBlk.append( vec_ij_index ) #build list of vector indices of each element of gateBlk mx
            for i,fi in enumerate(mapBlk):
                for j,fj in enumerate(mapBlk):
                    gateTermInStdBasis[fi,fj] = gateBlk[i,j]


        elif gateName in ('CX','CY','CZ'): #two-qubit gate names
            assert(len(args) == 3) # theta, qubit-label1, qubit-label2
            theta = eval( args[0], {"__builtins__":None}, {'pi': _np.pi})
            label1 = args[1]; assert(label1.startswith('Q'))
            label2 = args[2]; assert(label2.startswith('Q'))

            if gateName == 'CX': ex = -1j * theta*_bt.sigmax/2
            elif gateName == 'CY': ex = -1j * theta*_bt.sigmay/2
            elif gateName == 'CZ': ex = -1j * theta*_bt.sigmaz/2
            Utarget = _spl.expm(ex) # 2x2 unitary matrix operating on target qubit
            Ugate = _np.identity(4, 'complex'); Ugate[2:,2:] = Utarget #4x4 unitary matrix operating on isolated two-qubit space

            iTensorProdBlk = tensorBlkIndices[label1] # index of tensor product block (of state space) this bit label is part of
            assert( iTensorProdBlk == tensorBlkIndices[label2] ) #labels must be members of the same tensor product block
            cohBlk = stateSpaceLabels[iTensorProdBlk]
            basisInds = []
            for l in cohBlk:
                assert(l[0] in ('L','Q')) #should have been checked above
                if l.startswith('L'): basisInds.append([0])
                elif l.startswith('Q'): basisInds.append([0,1])

            tensorBlkBasis = list(_itertools.product(*basisInds))
            K1 = cohBlk.index(label1)
            K2 = cohBlk.index(label2)
            N = len(tensorBlkBasis)
            UcohBlk = _np.identity( N, 'complex' ) # unitary matrix operating on relevant tensor product block part of state
            for i,b1 in enumerate(tensorBlkBasis):
                for j,b2 in enumerate(tensorBlkBasis):
                    b1p = list(b1); del b1p[max(K1,K2)]; del b1p[min(K1,K2)] # b1' -- remove basis indices for tensor
                    b2p = list(b2); del b2p[max(K1,K2)]; del b2p[min(K1,K2)] # b2'      product parts we operate on
                    if b1p == b2p:   #if all parts of tensor product match except for qubits we're operating on
                        UcohBlk[i,j] = Ugate[ 2*b1[K1]+b1[K2], 2*b2[K1]+b2[K2] ] # then fill in element

            #print "UcohBlk = \n",UcohBlk

            UcohBlkc = UcohBlk.conjugate()
            gateBlk = _np.kron(UcohBlk,UcohBlkc) # N^2 x N^2 mx operating on vectorized tensor product block of densty matrix

            #Map gateBlk's basis into final gate basis
            mapBlk = []
            s = startIndex[iTensorProdBlk] #within state space (i.e. row or col of density matrix)
            cohBlkSize = UcohBlk.shape[0]
            for i in range(cohBlkSize):
                for j in range(cohBlkSize):
                    vec_ij_index = (s+i)*dmDim + (s+j) #vectorize by concatenating rows
                    mapBlk.append( vec_ij_index ) #build list of vector indices of each element of gateBlk mx
            for i,fi in enumerate(mapBlk):
                for j,fj in enumerate(mapBlk):
                    gateTermInStdBasis[fi,fj] = gateBlk[i,j]


        elif gateName == "LX":  #TODO - better way to describe leakage?
            assert(len(args) == 3) # theta, dmIndex1, dmIndex2 - X rotation between any two density matrix basis states
            theta = eval( args[0], {"__builtins__":None}, {'pi': _np.pi})
            i1 = int(args[1])
            i2 = int(args[2])
            ex = -1j * theta*_bt.sigmax/2
            Ugate = _spl.expm(ex) # 2x2 unitary matrix operating on the i1-th and i2-th states of the state space basis
            Utot = _np.identity(dmDim, 'complex')
            Utot[ i1,i1 ] = Ugate[0,0]
            Utot[ i1,i2 ] = Ugate[0,1]
            Utot[ i2,i1 ] = Ugate[1,0]
            Utot[ i2,i2 ] = Ugate[1,1]

            Utotc = Utot.conjugate()
            gateBlk = _np.kron(Utot,Utotc) # N^2 x N^2 mx operating on vectorized tensor product block of densty matrix

            #Map gateBlk's basis (vectorized 2x2) into final gate basis
            mapBlk = [] #note: "start index" is effectively zero since we're mapping all the blocs
            for i in range(dmDim):
                for j in range(dmDim):
                    vec_ij_index = (i)*dmDim + (j) #vectorize by concatenating rows
                    mapBlk.append( vec_ij_index ) #build list of vector indices of each element of gateBlk mx
            for i,fi in enumerate(mapBlk):
                for j,fj in enumerate(mapBlk):
                    gateTermInStdBasis[fi,fj] = gateBlk[i,j]

            #sq = startIndex[qbIndex]; sc = startIndex[clIndex]  #sq,sc are density matrix start indices
            #vsq = dmiToVi[ (sq,sq) ]; vsc = dmiToVi[ (sc,sc) ]  # vector indices of (sq,sq) and (sc,sc) density matrix elements
            #vsq1 = dmiToVi[ (sq,sq+1) ]; vsq2 = dmiToVi[ (sq+1,sq) ]  # vector indices of qubit coherences
            #
            ## action = swap (sq,sq) and (sc,sc) elements of a d.mx. and destroy coherences within qubit
            #gateTermInStdBasis[vsq,vsc] = gateTermInStdBasis[vsc,vsq] = 1.0
            #gateTermInStdBasis[vsq,vsq] = gateTermInStdBasis[vsc,vsc] = 0.0
            #gateTermInStdBasis[vsq1,vsq1] = gateTermInStdBasis[vsq2,vsq2] = 0.0


#        elif gateName == "Flip":
#            assert(len(args) == 2) # clevel-index0, clevel-index1
#            indx0 = int(args[0])
#            indx1 = int(args[1])
#            assert(indx0 != indx1)
#            assert(bitLabels[indx0] == 'L' and bitLabels[indx1] == 'L')
#
#            s0 = startIndex[indx0]; s1 = startIndex[indx1] #density matrix indices
#            vs0 = dmiToVi[ (s0,s0) ]; vs1 = dmiToVi[ (s1,s1) ]  # vector indices of (s0,s0) and (s1,s1) density matrix elements
#
#            # action = swap (s0,s0) and (s1,s1) elements of a d.mx.
#            gateTermInStdBasis[vs0,vs1] = gateTermInStdBasis[vs1,vs0] = 1.0
#            gateTermInStdBasis[vs0,vs0] = gateTermInStdBasis[vs1,vs1] = 0.0

        else: raise ValueError("Invalid gate name: %s" % gateName)

        gateInStdBasis = _np.dot(gateInStdBasis, gateTermInStdBasis)

    #Pare down gateInStdBasis to only include those matrix unit basis elements that are allowed to be nonzero
    gateInReducedStdBasis = _bt.contract_to_std_direct_sum_mx(gateInStdBasis, stateSpaceDims)

    #Change from std (mx unit) basis to another if requested
    gateMxInFinalBasis = _bt.change_basis(gateInReducedStdBasis, "std", basis, stateSpaceDims)
    
    return _gate.FullyParameterizedGate(gateMxInFinalBasis)


def build_gate(stateSpaceDims, stateSpaceLabels, gateExpr, basis="gm", parameterization="full", unitaryEmbedding=False):
#coherentStateSpaceBlockDims
    """
    Build a Gate object from an expression.

    Parameters
    ----------
    stateSpaceDims : list of ints
        Dimenstions specifying the structure of the density-matrix space.
        Elements correspond to block dimensions of an allowed density matrix in
        the standard basis, and the density-matrix space is the direct sum of
        linear spaces of dimension block-dimension^2.

    stateSpaceLabels : a list of tuples
        Each tuple corresponds to a block of a density matrix in the standard
        basis (and therefore a component of the direct-sum density matrix
        space). Elements of a tuple are user-defined labels beginning with "L"
        (single level) or "Q" (two-level; qubit) which interpret the
        d-dimensional state space corresponding to a d x d block as a tensor
        product between qubit and single level systems.

    gateExpr : string
        expression for the gate to build.  String is first split into parts
        delimited by the colon (:) character, which are composed together to
        create the final gate.  Each part takes on of the allowed forms:

        - I(ssl_0, ...) = identity operation on one or more state space labels
          (ssl_i)
        - X(theta, ssl) = x-rotation by theta radians of qubit labeled by ssl
        - Y(theta, ssl) = y-rotation by theta radians of qubit labeled by ssl
        - Z(theta, ssl) = z-rotation by theta radians of qubit labeled by ssl
        - CX(theta, ssl0, ssl1) = controlled x-rotation by theta radians.  Acts
          on qubit labeled by ssl1 with ssl0 being the control.
        - CY(theta, ssl0, ssl1) = controlled y-rotation by theta radians.  Acts
          on qubit labeled by ssl1 with ssl0 being the control.
        - CZ(theta, ssl0, ssl1) = controlled z-rotation by theta radians.  Acts
          on qubit labeled by ssl1 with ssl0 being the control.
        - LX(theta, i0, i1) = leakage between states i0 and i1.  Implemented as
          an x-rotation between states with integer indices i0 and i1 followed
          by complete decoherence between the states.

    basis : {'gm','pp','std','qt'}, optional
        the basis of the returned gate.

        - "std" = gate matrix operates on density mx expressed as sum of matrix
          units
        - "gm"  = gate matrix operates on dentity mx expressed as sum of
          normalized Gell-Mann matrices
        - "pp"  = gate matrix operates on density mx expressed as sum of
          tensor-product of Pauli matrices
        - "qt"  = gate matrix operates on density mx expressed as sum of
          Qutrit basis matrices

    parameterization : {"full","TP","static","linear","linearTP"}, optional
        How to parameterize the resulting gate.

        - "full" = return a FullyParameterizedGate.
        - "TP" = return a TPParameterizedGate.
        - "static" = return a StaticGate.
        - "linear" = if possible, return a LinearlyParameterizedGate that
          parameterizes only the pieces explicitly present in gateExpr.
        - "linearTP" = if possible, return a LinearlyParameterizedGate that
          parameterizes only the TP pieces explicitly present in gateExpr.

    unitaryEmbedding : bool, optional
        An interal switch determining how the gate is constructed.  Should have
        no bearing on the output except in determining how to parameterize a
        non-FullyParameterizedGate.  It's best to leave this to False unless
        you really know what you're doing.  Currently, only works for
        parameterization == 'full'.

    Returns
    -------
    Gate
        A gate object representing the gate given by gateExpr in the desired
        basis.
    """
    # gateExpr can contain single qubit ops: X(theta) ,Y(theta) ,Z(theta)
    #                      two qubit ops: CNOT
    #                      clevel qubit ops: Leak
    #                      two clevel opts: Flip
    #  each of which is given additional parameters specifying which indices it acts upon

    dmDim, gateDim, blockDims = _bt._processBlockDims(stateSpaceDims)
    #fullOpDim = dmDim**2

    #Store each tensor product blocks start index (within the density matrix), which tensor product block
    #  each label is in, and check to make sure dimensions match stateSpaceDims
    tensorBlkIndices = {}; startIndex = []; M = 0
    assert( len(blockDims) == len(stateSpaceLabels) )
    for k, blockDim in enumerate(blockDims):
        startIndex.append(M); M += blockDim

        #Make sure tensor-product interpretation agrees with given dimension
        tensorBlkDim = 1 #dimension of this coherent block of the *density matrix*
        for s in stateSpaceLabels[k]:
            tensorBlkIndices[s] = k
            if s.startswith('Q'): tensorBlkDim *= 2
            elif s.startswith('L'): tensorBlkDim *= 1
            else: raise ValueError("Invalid state space specifier: %s" % s)
        if tensorBlkDim != blockDim:
            raise ValueError("State labels %s for tensor product block %d have dimension %d != given dimension %d" \
                                 % (stateSpaceLabels[k], k, tensorBlkDim, blockDim))


    # ----------------------------------------------------------------------------------------------------------------------------------------
    # -- Helper Functions --------------------------------------------------------------------------------------------------------------------
    # ----------------------------------------------------------------------------------------------------------------------------------------

    def equals_except(list1, list2, exemptIndices):
        for i,(l1,l2) in enumerate(zip(list1,list2)):
            if i in exemptIndices: continue
            if l1 != l2: return False
        return True

    def embed_gate_unitary(Ugate, labels):  # Ugate should be in std basis (really no other basis it could be since gm and pp are only for acting on dm space)
        iTensorProdBlks = [ tensorBlkIndices[label] for label in labels ] # index of tensor product block (of state space) a bit label is part of
        if len(set(iTensorProdBlks)) > 1:
            raise ValueError("All qubit labels of a multi-qubit gate must correspond to the same tensor-product-block of the state space")

        iTensorProdBlk = iTensorProdBlks[0] #because they're all the same (tested above)
        tensorProdBlkLabels = stateSpaceLabels[iTensorProdBlk]
        basisInds = [] # list of *state* indices of each component of the tensor product block
        for l in tensorProdBlkLabels:
            assert(l[0] in ('L','Q')) #should have been checked above
            if l.startswith('L'): basisInds.append([0])
            elif l.startswith('Q'): basisInds.append([0,1])

        tensorBlkBasis = list(_itertools.product(*basisInds)) #state-space basis (remember tensor-prod-blocks are in state space)
        N = len(tensorBlkBasis) #size of state space (not density matrix space, which is N**2)

        labelIndices = [ tensorProdBlkLabels.index(label) for label in labels ]
        labelMultipliers = []; stateSpaceDim = 1
        for l in reversed(labels):
            labelMultipliers.append(stateSpaceDim)
            if l.startswith('L'): stateSpaceDim *= 1 #Warning? - having a gate operate on an L label doesn't really do anything...
            elif l.startswith('Q'): stateSpaceDim *= 2
        labelMultipliers.reverse() #reverse back to labels order (labels was reversed in loop above)
        labelMultipliers = _np.array(labelMultipliers,'i') #so we can use _np.dot below
        assert(stateSpaceDim == Ugate.shape[0] == Ugate.shape[1])

        # Unitary op approach: build unitary acting on state space than use kron => map acting on vec(density matrix) space
        UcohBlk = _np.identity( N, 'complex' ) # unitary matrix operating on relevant tensor product block part of state
        for i,b1 in enumerate(tensorBlkBasis):
            for j,b2 in enumerate(tensorBlkBasis):
                if equals_except(b1,b2,labelIndices): #if all parts of tensor prod match except for qubit(s) we're operating on
                    gate_b1 = _np.array([ b1[K] for K in labelIndices ],'i') #basis indices for just the qubits we're operating on
                    gate_b2 = _np.array([ b2[K] for K in labelIndices ],'i') # - i.e. those corresponding to the given Ugate
                    gate_i = _np.dot(labelMultipliers, gate_b1)
                    gate_j = _np.dot(labelMultipliers, gate_b2)
                    UcohBlk[i,j] = Ugate[ gate_i, gate_j ] # fill in element
                    #FUTURE: could keep track of what Ugate <-> UcohBlk elements for parameterization here

        UcohBlkc = UcohBlk.conjugate()
        gateBlk = _np.kron(UcohBlk,UcohBlkc) # N^2 x N^2 mx operating on vectorized tensor product block of densty matrix

        #print "DEBUG: Ugate = \n", Ugate
        #print "DEBUG: UcohBlk = \n", UcohBlk

        #Map gateBlk's basis into final gate basis (shift basis indices due to the composition of different direct-sum
        # blocks along diagonal of final gate mx)
        offset = sum( [ blockDims[i]**2 for i in range(0,iTensorProdBlk) ] ) #number of basis elements preceding our block's elements
        finalGateInStdBasis = _np.identity( gateDim, 'complex' )             # operates on entire state space (direct sum of tensor prod. blocks)
        finalGateInStdBasis[offset:offset+N**2,offset:offset+N**2] = gateBlk # gateBlk gets offset along diagonal by the numer of preceding basis elements

        if parameterization != "full":
            raise ValueError("Unitary embedding is only implemented for parmeterization='full'")

        finalGateInFinalBasis = _bt.change_basis(finalGateInStdBasis, "std", basis, blockDims)
        return _gate.FullyParameterizedGate(finalGateInFinalBasis)


    def embed_gate(gatemx, labels, indicesToParameterize="all"):
        #print "DEBUG: embed_gate gatemx = \n", gatemx
        iTensorProdBlks = [ tensorBlkIndices[label] for label in labels ] # index of tensor product block (of state space) a bit label is part of
        assert( len(set(iTensorProdBlks)) == 1 )
          #All qubit labels of a multi-qubit gate must correspond to the
          # same tensor-product-block of the state space -- checked previously

        iTensorProdBlk = iTensorProdBlks[0] #because they're all the same (tested above)
        tensorProdBlkLabels = stateSpaceLabels[iTensorProdBlk]
        basisInds = [] # list of possible *density-matrix-space* indices of each component of the tensor product block
        for l in tensorProdBlkLabels:
            assert(l[0] in ('L','Q')) #should have already been checked
            if l.startswith('L'): basisInds.append([0]) # I
            elif l.startswith('Q'): basisInds.append([0,1,2,3])  # I, X, Y, Z

        tensorBlkEls = list(_itertools.product(*basisInds)) #dm-space basis
        lookup_blkElIndex = { tuple(b):i for i,b in enumerate(tensorBlkEls) } # index within vec(tensor prod blk) of each basis el
        N = len(tensorBlkEls) #size of density matrix space
        assert( N == blockDims[iTensorProdBlk]**2 )

        # Gate matrix approach: insert elements of gatemx into map acting on vec(density matrix) space
        gateBlk = _np.identity( N, 'd' ) # matrix operating on vec(tensor product block), (tensor prod blk is a part of the total density mx)
          #Note: because we're in the Pauil-product basis this is a *real* matrix (and gatemx should have only real elements and be in the pp basis)

        # Separate the components of the tensor product that are not operated on, i.e. that our final map just acts as identity w.r.t.
        basisInds_noop = basisInds[:]
        labelIndices = [ tensorProdBlkLabels.index(label) for label in labels ]
        for labelIndex in sorted(labelIndices,reverse=True):
            del basisInds_noop[labelIndex]
        tensorBlkEls_noop = list(_itertools.product(*basisInds_noop)) #dm-space basis for noop-indices only
        parameterToBaseIndicesMap = {}

        def decomp_gate_index(indx):
            """ Decompose index of a Pauli-product matrix into indices of each
            Pauli in the product """
            ret = []; divisor = 1; divisors = []
            #print "Decomp %d" % indx,
            for l in labels:
                divisors.append(divisor)
                if l.startswith('Q'): divisor *= 4
                elif l.startswith('L'): divisor *= 1
            for d in reversed(divisors):
                ret.append( indx // d )
                indx = indx % d
            #print " => %s (div = %s)" % (str(ret), str(divisors))
            return ret

        def merge_gate_and_noop_bases(gate_b, noop_b):
            """
            Merge the Pauli basis indices for the "gate"-parts of the total
            basis contained in gate_b (i.e. of the components of the tensor
            product space that are operated on) and the "noop"-parts contained
            in noop_b.  Thus, len(gate_b) + len(noop_b) == len(basisInds), and
            this function merges together basis indices for the operated-on and
            not-operated-on tensor product components.
            Note: return value always have length == len(basisInds) == number
            of componens
            """
            ret = list(noop_b[:])    #start with noop part...
            for li,b_el in sorted( zip(labelIndices,gate_b), key=lambda x: x[0]):
                ret.insert(li, b_el) #... and insert gate parts at proper points
            return ret


        for gate_i in range(gatemx.shape[0]):     # rows ~ "output" of the gate map
            for gate_j in range(gatemx.shape[1]): # cols ~ "input"  of the gate map
                if indicesToParameterize == "all":
                    iParam = gate_i*gatemx.shape[1] + gate_j #index of (i,j) gate parameter in 1D array of parameters (flatten gatemx)
                    parameterToBaseIndicesMap[ iParam ] = []
                elif indicesToParameterize == "TP":
                    if gate_i > 0:
                        iParam = (gate_i-1)*gatemx.shape[1] + gate_j
                        parameterToBaseIndicesMap[ iParam ] = []
                    else:
                        iParam = None
                elif (gate_i,gate_j) in indicesToParameterize:
                    iParam = indicesToParameterize.index( (gate_i,gate_j) )
                    parameterToBaseIndicesMap[ iParam ] = []
                else:
                    iParam = None #so we don't parameterize below

                gate_b1 = decomp_gate_index(gate_i) # gate_b? are lists of dm basis indices, one index per
                gate_b2 = decomp_gate_index(gate_j) #  tensor product component that the gate operates on (2 components for a 2-qubit gate)

                for b_noop in tensorBlkEls_noop: #loop over all state configurations we don't operate on - so really a loop over diagonal dm elements
                    b_out = merge_gate_and_noop_bases(gate_b1, b_noop)  # using same b_noop for in and out says we're acting
                    b_in  = merge_gate_and_noop_bases(gate_b2, b_noop)  #  as the identity on the no-op state space
                    out_vec_index = lookup_blkElIndex[ tuple(b_out) ] # index of output dm basis el within vec(tensor block basis)
                    in_vec_index  = lookup_blkElIndex[ tuple(b_in) ]  # index of input dm basis el within vec(tensor block basis)

                    gateBlk[ out_vec_index, in_vec_index ] = gatemx[ gate_i, gate_j ]
                    if iParam is not None:
                        # keep track of what gateBlk <-> gatemx elements for parameterization
                        parameterToBaseIndicesMap[ iParam ].append( (out_vec_index, in_vec_index) )


        #Map gateBlk's basis into final gate basis (shift basis indices due to the composition of different direct-sum
        # blocks along diagonal of final gate mx)
        offset = sum( [ blockDims[i]**2 for i in range(0,iTensorProdBlk) ] ) #number of basis elements preceding our block's elements
        finalGate = _np.identity( gateDim, 'd' )              # operates on entire state space (direct sum of tensor prod. blocks)
        finalGate[offset:offset+N,offset:offset+N] = gateBlk  # gateBlk gets offset along diagonal by the number of preceding basis elements
        # Note: final is a *real* matrix whose basis is the pauli-product basis in the iTensorProdBlk-th block, concatenated with
        #   bases for the other blocks - say the "std" basis (which one does't matter since the identity is the same for std, gm, and pp)

        #print "DEBUG: embed_gate gateBlk = \n", gateBlk
        full_finalToPP = _np.identity( gateDim, 'complex' )
        full_ppToFinal = _np.identity( gateDim, 'complex' )

        if basis == "std":
            ppToStd = _bt.pp_to_std_transform_matrix(blockDims[iTensorProdBlk]); stdToPP = _np.linalg.inv(ppToStd)
            full_ppToFinal[offset:offset+N,offset:offset+N] = ppToStd
            full_finalToPP[offset:offset+N,offset:offset+N] = stdToPP
            realMx = False

        elif basis == "gm":
            ppToStd = _bt.pp_to_std_transform_matrix(blockDims[iTensorProdBlk]); stdToPP = _np.linalg.inv(ppToStd)
            gmToStd = _bt.gm_to_std_transform_matrix(blockDims[iTensorProdBlk]); stdToGM = _np.linalg.inv(gmToStd)
            full_ppToFinal[offset:offset+N,offset:offset+N] = _np.dot( stdToGM, ppToStd )
            full_finalToPP[offset:offset+N,offset:offset+N] = _np.dot( stdToPP, gmToStd )
            realMx = True

        elif basis == "pp":
            # Note: finalGate may have some non-power-of-2 dimensional blocks that can't be in a "pp" basis,
            #  but don't check for this here as they'll be caught by any gate term that operates on that block
            realMx = True

        elif basis == "qt":
            ppToStd = _bt.pp_to_std_transform_matrix(blockDims[iTensorProdBlk]); stdToPP = _np.linalg.inv(ppToStd)
            qtToStd = _bt.qt_to_std_transform_matrix(blockDims[iTensorProdBlk]); stdToQT = _np.linalg.inv(qtToStd)
            full_ppToFinal[offset:offset+N,offset:offset+N] = _np.dot( stdToQT, ppToStd )
            full_finalToPP[offset:offset+N,offset:offset+N] = _np.dot( stdToPP, qtToStd )
            realMx = True

        else: raise ValueError("Invalid 'basis' parameter: %s (must by 'std', 'gm', 'pp', or 'qt')" % basis)

        if parameterization == "full":
            finalGateInFinalBasis = _np.dot(full_ppToFinal,
                                            _np.dot( finalGate, full_finalToPP))
            return _gate.FullyParameterizedGate(
                _np.real(finalGateInFinalBasis)
                if realMx else finalGateInFinalBasis )

        if parameterization == "static":
            finalGateInFinalBasis = _np.dot(full_ppToFinal,
                                            _np.dot( finalGate, full_finalToPP))
            return _gate.StaticGate(
                _np.real(finalGateInFinalBasis)
                if realMx else finalGateInFinalBasis )

        if parameterization == "TP":
            finalGateInFinalBasis = _np.dot(full_ppToFinal,
                                            _np.dot( finalGate, full_finalToPP))
            if not realMx:
                raise ValueError("TP gates must be real. Failed to build gate!")
            return _gate.TPParameterizedGate(_np.real(finalGateInFinalBasis))

        elif parameterization in ("linear","linearTP"):
            #OLD (INCORRECT) -- but could give this as paramArray if gave zeros as base matrix instead of finalGate
            # paramArray = gatemx.flatten() if indicesToParameterize == "all" else _np.array([gatemx[t] for t in indicesToParameterize])

            #Set all params to *zero* since base matrix contains all initial elements -- parameters just give deviation
            if indicesToParameterize == "all":
                paramArray = _np.zeros(gatemx.size, 'd')
            elif indicesToParameterize == "TP":
                paramArray = _np.zeros(gatemx.size - gatemx.shape[1], 'd')
            else:
                paramArray = _np.zeros(len(indicesToParameterize), 'd' )

            return _gate.LinearlyParameterizedGate(
                finalGate, paramArray, parameterToBaseIndicesMap,
                full_ppToFinal, full_finalToPP, realMx )


        else:
            raise ValueError("Invalid 'parameterization' parameter: " +
                             "%s (must by 'full', 'TP', 'static', 'linear' or 'linearTP')"
                             % parameterization)


    # ----------------------------------------------------------------------------------------------------------------------------------------
    # -- End Helper Functions ----------------------------------------------------------------------------------------------------------------
    # ----------------------------------------------------------------------------------------------------------------------------------------


    #print "DB: dim = ",dim, " dmDim = ",dmDim
    gateInFinalBasis = None #what will become the final gate matrix
    defaultI2P = "all" if parameterization != "linearTP" else "TP"
      #default indices to parameterize (I2P) - used only when 
      # creating parameterized gates

    exprTerms = gateExpr.split(':')
    for exprTerm in exprTerms:

        l = exprTerm.index('('); r = exprTerm.rindex(')')
        gateName = exprTerm[0:l]
        argsStr = exprTerm[l+1:r]
        args = argsStr.split(',')

        if gateName == "I":
            labels = args # qubit labels (TODO: what about 'L' labels? -- not sure if they work with this...)
            stateSpaceDim = 1
            for l in labels:
                if l.startswith('Q'): stateSpaceDim *= 2
                elif l.startswith('L'): stateSpaceDim *= 1
                else: raise ValueError("Invalid state space label: %s" % l)

            if unitaryEmbedding:
                Ugate = _np.identity(stateSpaceDim, 'complex') #complex because in std state space basis
                gateTermInFinalBasis = embed_gate_unitary(Ugate, tuple(labels)) #Ugate assumed to be in std basis (really the only option)
            else:
                pp_gateMx = _np.identity(stateSpaceDim**2, 'd') # *real* 4x4 mx in Pauli-product basis -- still just the identity!
                gateTermInFinalBasis = embed_gate(pp_gateMx, tuple(labels), defaultI2P) # pp_gateMx assumed to be in the Pauli-product basis

        elif gateName == "D":  #like 'I', but only parameterize the diagonal elements - so can be a depolarization-type map
            labels = args # qubit labels (TODO: what about 'L' labels? -- not sure if they work with this...)
            stateSpaceDim = 1
            for l in labels:
                if l.startswith('Q'): stateSpaceDim *= 2
                elif l.startswith('L'): stateSpaceDim *= 1
                else: raise ValueError("Invalid state space label: %s" % l)

            if unitaryEmbedding or parameterization not in ("linear","linearTP"):
                raise ValueError("'D' gate only makes sense to use when unitaryEmbedding is False and parameterization == 'linear'")

            if defaultI2P == "TP":
                indicesToParameterize = [ (i,i) for i in range(1,stateSpaceDim**2) ] #parameterize only the diagonals els after the first
            else:
                indicesToParameterize = [ (i,i) for i in range(0,stateSpaceDim**2) ] #parameterize only the diagonals els
            pp_gateMx = _np.identity(stateSpaceDim**2, 'd') # *real* 4x4 mx in Pauli-product basis -- still just the identity!
            gateTermInFinalBasis = embed_gate(pp_gateMx, tuple(labels), indicesToParameterize) # pp_gateMx assumed to be in the Pauli-product basis


        elif gateName in ('X','Y','Z'): #single-qubit gate names
            assert(len(args) == 2) # theta, qubit-index
            theta = eval( args[0], {"__builtins__":None}, {'pi': _np.pi})
            label = args[1].strip(); assert(label.startswith('Q'))

            if gateName == 'X': ex = -1j * theta*_bt.sigmax/2
            elif gateName == 'Y': ex = -1j * theta*_bt.sigmay/2
            elif gateName == 'Z': ex = -1j * theta*_bt.sigmaz/2

            Ugate = _spl.expm(ex) # 2x2 unitary matrix operating on single qubit in [0,1] basis
            if unitaryEmbedding:
                gateTermInFinalBasis = embed_gate_unitary(Ugate, (label,)) #Ugate assumed to be in std basis (really the only option)
            else:
                Ugatec = Ugate.conjugate()
                gateMx = _np.kron(Ugate,Ugatec) # complex 4x4 mx operating on vectorized 1Q densty matrix in std basis
                pp_gateMx = _bt.std_to_pp(gateMx) # *real* 4x4 mx in Pauli-product basis -- better for parameterization
                gateTermInFinalBasis = embed_gate(pp_gateMx, (label,), defaultI2P) # pp_gateMx assumed to be in the Pauli-product basis

        elif gateName == 'N': #more general single-qubit gate
            assert(len(args) == 5) # theta, sigmaX-coeff, sigmaY-coeff, sigmaZ-coeff, qubit-index
            theta = eval( args[0], {"__builtins__":None}, {'pi': _np.pi, 'sqrt': _np.sqrt})
            sxCoeff = eval( args[1], {"__builtins__":None}, {'pi': _np.pi, 'sqrt': _np.sqrt})
            syCoeff = eval( args[2], {"__builtins__":None}, {'pi': _np.pi, 'sqrt': _np.sqrt})
            szCoeff = eval( args[3], {"__builtins__":None}, {'pi': _np.pi, 'sqrt': _np.sqrt})
            label = args[4].strip(); assert(label.startswith('Q'))

            ex = -1j * theta * ( sxCoeff * _bt.sigmax/2. + syCoeff * _bt.sigmay/2. + szCoeff * _bt.sigmaz/2.)
            Ugate = _spl.expm(ex) # 2x2 unitary matrix operating on single qubit in [0,1] basis
            if unitaryEmbedding:
                gateTermInFinalBasis = embed_gate_unitary(Ugate, (label,)) #Ugate assumed to be in std basis (really the only option)
            else:
                Ugatec = Ugate.conjugate()
                gateMx = _np.kron(Ugate,Ugatec) # complex 4x4 mx operating on vectorized 1Q densty matrix in std basis
                pp_gateMx = _bt.std_to_pp(gateMx) # *real* 4x4 mx in Pauli-product basis -- better for parameterization
                gateTermInFinalBasis = embed_gate(pp_gateMx, (label,), defaultI2P) # pp_gateMx assumed to be in the Pauli-product basis

        elif gateName in ('CX','CY','CZ'): #two-qubit gate names
            assert(len(args) == 3) # theta, qubit-label1, qubit-label2
            theta = eval( args[0], {"__builtins__":None}, {'pi': _np.pi})
            label1 = args[1].strip(); assert(label1.startswith('Q'))
            label2 = args[2].strip(); assert(label2.startswith('Q'))

            if gateName == 'CX': ex = -1j * theta*_bt.sigmax/2.
            elif gateName == 'CY': ex = -1j * theta*_bt.sigmay/2.
            elif gateName == 'CZ': ex = -1j * theta*_bt.sigmaz/2.
            Utarget = _spl.expm(ex) # 2x2 unitary matrix operating on target qubit
            Ugate = _np.identity(4, 'complex'); Ugate[2:,2:] = Utarget #4x4 unitary matrix operating on isolated two-qubit space

            if unitaryEmbedding:
                gateTermInFinalBasis = embed_gate_unitary(Ugate, (label1,label2)) #Ugate assumed to be in std basis (really the only option)
            else:
                Ugatec = Ugate.conjugate()
                gateMx = _np.kron(Ugate,Ugatec) # complex 16x16 mx operating on vectorized 2Q densty matrix in std basis
                pp_gateMx = _bt.std_to_pp(gateMx) # *real* 16x16 mx in Pauli-product basis -- better for parameterization
                gateTermInFinalBasis = embed_gate(pp_gateMx, (label1,label2), defaultI2P) # pp_gateMx assumed to be in the Pauli-product basis

        elif gateName == "LX":  #TODO - better way to describe leakage?
            assert(len(args) == 3) # theta, dmIndex1, dmIndex2 - X rotation between any two density matrix basis states
            theta = eval( args[0], {"__builtins__":None}, {'pi': _np.pi})
            i1 = int(args[1])  # row/column index of a single *state* within the density matrix
            i2 = int(args[2])  # row/column index of a single *state* within the density matrix
            ex = -1j * theta*_bt.sigmax/2
            Ugate = _spl.expm(ex) # 2x2 unitary matrix operating on the i1-th and i2-th states of the state space basis

            Utot = _np.identity(dmDim, 'complex')
            Utot[ i1,i1 ] = Ugate[0,0]
            Utot[ i1,i2 ] = Ugate[0,1]
            Utot[ i2,i1 ] = Ugate[1,0]
            Utot[ i2,i2 ] = Ugate[1,1]
            Utotc = Utot.conjugate()
            gateTermInStdBasis = _np.kron(Utot,Utotc) # dmDim^2 x dmDim^2 mx operating on vectorized total densty matrix
            gateTermInReducedStdBasis = _bt.contract_to_std_direct_sum_mx(gateTermInStdBasis, blockDims)

            gateMxInFinalBasis = _bt.change_basis(gateTermInReducedStdBasis, "std", basis, blockDims)
            gateTermInFinalBasis = _gate.FullyParameterizedGate(gateMxInFinalBasis)

        else: raise ValueError("Invalid gate name: %s" % gateName)

        if gateInFinalBasis is None:
            gateInFinalBasis = gateTermInFinalBasis
        else:
            gateInFinalBasis = _gate.compose( gateInFinalBasis, gateTermInFinalBasis )

    return gateInFinalBasis # a Gate object





def build_gateset(stateSpaceDims, stateSpaceLabels,
                  gateLabels, gateExpressions,
                  prepLabels, prepExpressions,
                  effectLabels, effectExpressions,
                  spamdefs, basis="gm", parameterization="full"):
    """
    Build a new GateSet given lists of gate labels and expressions.

    Parameters
    ----------
    stateSpaceDims : list of ints
       Dimensions specifying the structure of the density-matrix space.
        Elements correspond to block dimensions of an allowed density matrix in
        the standard basis, and the density-matrix space is the direct sum of
        linear spaces of dimension block-dimension^2.

    stateSpaceLabels : a list of tuples
        Each tuple corresponds to a block of a density matrix in the standard
        basis (and therefore a component of the direct-sum density matrix
        space). Elements of a tuple are user-defined labels beginning with "L"
        (single level) or "Q" (two-level; qubit) which interpret the
        d-dimensional state space corresponding to a d x d block as a tensor
        product between qubit and single level systems.

    gateLabels : list of strings
       A list of labels for each created gate in the final gateset.  To
        conform with text file parsing conventions these names should begin
        with a capital G and can be followed by any number of lowercase
        characters, numbers, or the underscore character.

    gateExpressions : list of strings
        A list of gate expressions, each corresponding to a gate label in
        gateLabels, which determine what operation each gate performs (see
        documentation for :meth:`build_gate`).

    prepLabels : list of string
        A list of labels for each created state preparation in the final
        gateset.  To conform with conventions these labels should begin with
        "rho".

    prepExpressions : list of strings
        A list of vector expressions for each state preparation vector (see
        documentation for :meth:`build_vector`).

    effectLabels : list of string
        A list of labels for each created and *parameterized* POVM effect in
        the final gateset.  To conform with conventions these labels should
        begin with "E".

    effectExpressions : list of strings
        A list of vector expressions for each POVM effect vector (see
        documentation for :meth:`build_vector`).

    spamdefs : dict
       A dictionary mapping spam labels to (prepLabel,ELabel) 2-tuples
        associating a particular state preparation and effect vector with a
        label.  prepLabel and ELabel must be contained in prepLabels and
        effectLabels respectively except for two special cases:

        1.  ELabel can be set to "remainder" to mean an effect vector that is
            the identity - (other effect vectors)
        2.  ELabel and prepLabel can both be "remainder" to mean a spam label
            that generates probabilities that are 1.0 - (sum of probabilities
            from all other spam labels).

    basis : {'gm','pp','std','qt'}, optional
        the basis of the matrices in the returned GateSet

        - "std" = gate matrix operates on density mx expressed as sum of matrix
          units
        - "gm"  = gate matrix operates on dentity mx expressed as sum of
          normalized Gell-Mann matrices
        - "pp"  = gate matrix operates on density mx expresses as sum of
          tensor-product of Pauli matrices
        - "qt"  = gate matrix operates on density mx expressed as sum of
          Qutrit basis matrices

    parameterization : {"full","TP","linear","linearTP"}, optional
        How to parameterize the gates of the resulting GateSet (see
        documentation for :meth:`build_gate`).

    Returns
    -------
    GateSet
        The created gate set.
    """
    defP = "TP" if (parameterization in ("TP","linearTP")) else "full"
    ret = _gateset.GateSet(default_param=defP)
                 #prep_prefix="rho", effect_prefix="E", gate_prefix="G",
                 #remainder_label="remainder", identity_label="identity")

    for label,rhoExpr in zip(prepLabels, prepExpressions):
        ret.preps[label] = build_vector(stateSpaceDims, stateSpaceLabels, rhoExpr, basis)
    for label,EExpr in zip(effectLabels,effectExpressions):
        ret.effects[label] = build_vector(stateSpaceDims, stateSpaceLabels, EExpr, basis)

    ret.povm_identity = build_identity_vec(stateSpaceDims, basis)

    #Note: since a GateSet's spamdefs are an *ordered* dictionary (for correspondence
    #  to row indices in some bulk_ operations), we need to set the spamdefs in a
    #  deterministic order -- e.g. we *cannot* iterate over the keys in a standard
    #  dictionary.  So, unless we're given an ordered dict, add the keys (spam labels)
    #  in alphabetical order.
    if isinstance(spamdefs, _collections.OrderedDict):
        for spamlabel,(rhoLbl,ELbl) in spamdefs.items():
            ret.spamdefs[spamlabel] = (rhoLbl,ELbl)
    else:
        for spamlabel in sorted(list(spamdefs.keys())):
            (rhoLbl,ELbl) = spamdefs[spamlabel]
            ret.spamdefs[spamlabel] = (rhoLbl,ELbl)

    for (gateLabel,gateExpr) in zip(gateLabels, gateExpressions):
        ret.gates[gateLabel] = build_gate(stateSpaceDims, stateSpaceLabels,
                                          gateExpr, basis, parameterization)

    if len(stateSpaceDims) == 1:
        basisDims = stateSpaceDims[0]
    else:
        basisDims = stateSpaceDims

    ret.set_basis(basis, basisDims)

    if parameterization == "full":
        ret.default_gauge_group = _gg.FullGaugeGroup(ret.dim)
    elif parameterization == "TP":
        ret.default_gauge_group = _gg.TPGaugeGroup(ret.dim)
    else:
        ret.default_gauge_group = None #assume no gauge freedom

    return ret


def build_alias_gateset(gs_primitives,alias_dict):
    """
    Creates a new gateset by composing the gates of an existing `GateSet`,
    `gs_primitives`, according to a dictionary of `GateString`s, `alias_dict`.
    The keys of `alias_dict` are the gate labels of the returned `GateSet1.
    SPAM vectors are unaltered, and simply copied from `gs_primitives`.

    Parameters
    ----------
    gs_primitives : GateSet
        A Gateset containing the "primitive" gates (those used to compose
        the gates of the returned gateset).
    
    alias_dict : dictionary
        A dictionary whose keys are strings and values are GateString objects
        specifying sequences of primitive gates.  Each key,value pair specifies
        the composition rule for a creating a gate in the returned gate set.
    
    Returns
    -------
    GateSet
        A gate set whose gates are compositions of primitive gates and whose
        spam operations are the same as those of `gs_primitives`.
    """
    gs_new = gs_primitives.copy()
    for gl in gs_primitives.gates.keys():
        del gs_new.gates[gl] #remove all gates from gs_new

    for gl,gstr in alias_dict.items():
        gs_new.gates[gl] = gs_primitives.product(gstr)
          #Creates fully parameterized gates by default...
    return gs_new




###SCRATCH
# Old from embed_gate:
#        for gate_i in range(gatemx.shape[0]):     # rows ~ "output" of the gate map
#            for gate_j in range(gatemx.shape[1]): # cols ~ "input"  of the gate map
#                gate_b1_ket, gate_b2_bra = decomp_gate_index(gate_i) # gate_b* are lists of state indices, one index per
#                gate_b2_ket, gate_b2_bra = decomp_gate_index(gate_j) #  tensor product component that the gate operates on (2 components for a 2-qubit gate)
#
#                for i,b_noop in enumerate(tensorBlkBasis_noop): #loop over all state configurations we don't operate on - so really a loop over diagonal dm elements
#
#                    out_ket = insert_gate_basis(gate_b1_ket, b_noop)  # using same b_noop for ket & bra says we're acting
#                    out_bra = insert_gate_basis(gate_b1_bra, b_noop)  #  as the identity on the no-op state space
#                    out_blkBasis_i = lookup_blkBasisIndex[ tuple(out_ket) ] # row index of ket within tensor block basis (state space basis)
#                    out_blkBasis_j = lookup_blkBasisIndex[ tuple(out_bra) ] # col index of ket within tensor block basis (state space basis)
#                    out_vec_index = N*out_blkBasis_i + out_blkBasis_j  # index of (row,col) tensor block element (vectorized row,col)
#
#                    in_ket = insert_gate_basis(gate_b2_ket, b_noop)  # using same b_noop for ket & bra says we're acting
#                    in_bra = insert_gate_basis(gate_b2_bra, b_noop)  #  as the identity on the no-op state space
#                    in_blkBasis_i = lookup_blkBasisIndex[ tuple(in_ket) ] # row index of ket within tensor block basis (state space basis)
#                    in_blkBasis_j = lookup_blkBasisIndex[ tuple(in_bra) ] # col index of ket within tensor block basis (state space basis)
#                    in_vec_index = N*in_blkBasis_i + in_blkBasis_j  # index of (row,col) tensor block element (vectorized row,col)
#
#                    gateBlk[ out_vec_index, in_vec_index ] = gatemx[ gate_i, gate_j ]



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Utility functions for creating and acting on lists of gate strings."""

import itertools as _itertools
import numpy as _np
import numpy.random as _rndm

from ..tools import listtools as _lt
from ..objects import gatestring as _gs
from .spamspecconstruction import get_spam_strs as _get_spam_strs


def _runExpression(str_expression, myLocals):
    exec( "result = " + str_expression, {"__builtins__": None}, myLocals )
    return myLocals.get("result",None)

def create_gatestring_list(*args,**kwargs):
    """
    Create a list of gate strings using a nested loop.  Positional arguments
    specify evaluation strings, which are evaluated within the inner-loop
    for a nested loop over all list or tuple type keyword arguments.

    Parameters
    ----------
    args : list of strings
        Positional arguments are strings that python can evaluate into either
        a tuple of gate labels or a GateString instance.  If evaluation raises
        an AssertionError (an assert statement fails) then that inner loop
        evaluation is skipped and list construction proceeds.

    kwargs : dict
        keys specify variable names that can be used in positional argument
        strings.

    Returns
    -------
    list of GateString

    Examples
    --------
    >>> from pygsti.construction import create_gatestring_list
    >>> As = [('a1',), ('a2',)]
    >>> Bs = [('b1',), ('b2',)]
    >>> list1 = create_gatestring_list('a', 'a+b', a=As, b=Bs)
    >>> print(list(map(str, list1)))
    ['a1', 'a2', 'a1b1', 'a1b2', 'a2b1', 'a2b2']

    You can change the order in which the different iterables are advanced.

    >>> list2 = create_gatestring_list('a+b', a=As, b=Bs, order=['a', 'b'])
    >>> print(list(map(str, list2)))
    ['a1b1', 'a1b2', 'a2b1', 'a2b2']
    >>> list3 = create_gatestring_list('a+b', a=As, b=Bs, order=['b', 'a'])
    >>> print(list(map(str, list3)))
    ['a1b1', 'a2b1', 'a1b2', 'a2b2']


    """
    lst = []

    loopOrder = kwargs.pop('order',[])
    loopLists = {}; loopLocals = { 'True': True, 'False': False, 'str':str, 'int': int, 'float': float}
    for key,val in kwargs.items():
        if type(val) in (list,tuple): #key describes a variable to loop over
            loopLists[key] = val
            if key not in loopOrder:
                loopOrder.append(key)
        else: # callable(val): #key describes a function or variable to pass through to exec
            loopLocals[key] = val

    #print "DEBUG: looplists = ",loopLists
    for str_expression in args:
        if len(str_expression) == 0:
            lst.append( _gs.GateString( () ) ); continue #special case

        keysToLoop = [ key for key in loopOrder if key in str_expression ]
        loopListsToLoop = [ loopLists[key] for key in keysToLoop ] #list of lists
        for allVals in _itertools.product(*loopListsToLoop):
            myLocals = { key:allVals[i] for i,key in enumerate(keysToLoop) }
            myLocals.update( loopLocals )
            try:
                result = _runExpression(str_expression, myLocals)
            except AssertionError: continue #just don't append

            if isinstance(result,_gs.GateString):
                gateStr = result
            elif isinstance(result,list) or isinstance(result,tuple):
                gateStr = _gs.GateString(result)
            elif isinstance(result,str):
                gateStr = _gs.GateString(None, result)
            lst.append(gateStr)

    return lst


def repeat(x,nTimes,assertAtLeastOneRep=False):
    """
    Repeat x nTimes times.

    Parameters
    ----------
    x : tuple or GateString
       the gate string to repeat

    nTimes : int
       the number of times to repeat x

    assertAtLeastOneRep : bool, optional
       if True, assert that nTimes > 0.  This can be useful when used
       within a create_gatestring_list inner loop to build a gate string
       lists where a string must be repeated at least once to be added
       to the list.

    Returns
    -------
    tuple or GateString (whichever x was)
    """
    if assertAtLeastOneRep:  assert(nTimes > 0)
    return x*nTimes

def repeat_count_with_max_length(x,maxLength,assertAtLeastOneRep=False):
    """
    Compute the number of times a gate string x must be repeated such that
    the repeated string has length <= maxLength.

    Parameters
    ----------
    x : tuple or GateString
       the gate string to repeat

    maxLength : int
       the maximum length

    assertAtLeastOneRep : bool, optional
       if True, assert that number of repetitions is > 0.
       This can be useful when used within a create_gatestring_list inner loop
       to build a gate string lists where a string must be repeated at
       least once to be added to the list.

    Returns
    -------
    int
      the number of repetitions.
    """
    l = len(x)
    if assertAtLeastOneRep:  assert(l <= maxLength)
    reps = maxLength//l if l > 0 else 0
    return reps

def repeat_with_max_length(x,maxLength,assertAtLeastOneRep=False):
    """
    Repeat the gate string x an integer number of times such that
    the repeated string has length <= maxLength.

    Parameters
    ----------
    x : tuple or GateString
       the gate string to repeat.

    maxLength : int
       the maximum length.

    assertAtLeastOneRep : bool, optional
       if True, assert that number of repetitions is > 0.
       This can be useful when used within a create_gatestring_list inner loop
       to build a gate string lists where a string must be repeated at
       least once to be added to the list.

    Returns
    -------
    tuple or GateString (whichever x was)
        the repeated gate string
    """
    return repeat(x,repeat_count_with_max_length(x,maxLength,assertAtLeastOneRep),assertAtLeastOneRep)

#Useful for anything?
#def repeat_empty(x,maxLength,assertAtLeastOneRep=False):
#    return ()

def repeat_and_truncate(x,N,assertAtLeastOneRep=False):
    """
    Repeat the gate string x so the repeated string has length greater than N,
    then truncate the string to be exactly length N.

    Parameters
    ----------
    x : tuple or GateString
       the gate string to repeat & truncate.

    N : int
       the truncation length.

    assertAtLeastOneRep : bool, optional
       if True, assert that number of repetitions is > 0.
       This is always the case when x has length > 0.

    Returns
    -------
    tuple or GateString (whichever x was)
        the repeated-then-truncated gate string
    """
    reps = repeat_count_with_max_length(x,N,assertAtLeastOneRep) + 1
    return (x*reps)[0:N]

def repeat_remainder_for_truncation(x,N,assertAtLeastOneRep=False):
    """
    Repeat the gate string x the fewest number of times such that the repeated
    string has length greater than or equal to N.  Return the portion of this
    repeated string from the N-th position to the end. Note that this corresponds
    to what is truncated in a call to repeateAndTruncate(x,N,assertAtLeastOneRep).

    Parameters
    ----------
    x : tuple or GateString
       the gate string to operate on.

    N : int
       the truncation length.

    assertAtLeastOneRep : bool, optional
       if True, assert that number of repetitions is > 0.
       This is always the case when x has length > 0.

    Returns
    -------
    tuple or GateString (whichever x was)
        the remainder gate string

    """
    reps = repeat_count_with_max_length(x,N,assertAtLeastOneRep)
    return x[0:(N - reps*len(x))]


def simplify_str(gateStringStr):
    """
    Simplify a string representation of a gate string.  The simplified
      string should evaluate to the same gate label tuple as the original.

    Parameters
    ----------
    gateStringStr : string
        the string representation of a gate string to be simplified.
        (e.g. "Gx{}", "Gy^1Gx")

    Returns
    -------
    string
        the simplified string representation.
    """
    s = gateStringStr.replace("{}","")
    s = s.replace("^1G","G")
    s = s.replace("^1(","(")
    s = s.replace("^1{","{")
    if s.endswith("^1"): s = s[:-2]
    return s if len(s) > 0 else "{}"



## gate-label-tuple function.  TODO: check if these are still needed.

def list_all_gatestrings(gateLabels, minlength, maxlength):
    """
    List all the gate strings in a given length range.

    Parameters
    ----------
    gateLabels : tuple
        tuple of gate labels to include in gate strings.

    minlength : int
        the minimum gate string length to return

    maxlength : int
        the maximum gate string length to return

    Returns
    -------
    list
        A list of GateString objects.
    """
    gateTuples = _itertools.chain(*[_itertools.product(gateLabels, repeat=N)
                                    for N in range(minlength, maxlength + 1)])
    return list(map(_gs.GateString, gateTuples))

def gen_all_gatestrings(gateLabels, minlength, maxlength):
    """ Generator version of list_all_gatestrings """
    gateTuples = _itertools.chain(*[_itertools.product(gateLabels, repeat=N)
                                    for N in range(minlength, maxlength + 1)])
    for gateTuple in gateTuples:
        yield _gs.GateString(gateTuple)

def list_all_gatestrings_onelen(gateLabels, length):
    """
    List all the gate strings of a given length.

    Parameters
    ----------
    gateLabels : tuple
        tuple of gate labels to include in gate strings.

    length : int
        the gate string length

    Returns
    -------
    list
        A list of GateString objects.
    """
    gateTuples = _itertools.product(gateLabels, repeat=length)
    return list(map(_gs.GateString, gateTuples))


def gen_all_gatestrings_onelen(gateLabels, length):
    """Generator version of list_all_gatestrings_onelen"""
#    OLD
#    if length == 0: yield _gs.GateString( () )
#    elif length == 1:
#        for g in gateLabels:
#            yield _gs.GateString( (g,) )
#    else:
#        for g in gateLabels:
#            for s in gen_all_gatestrings_onelen(gateLabels, length-1):
#                yield _gs.GateString( (g,) ) + s
    for gateTuple in _itertools.product(gateLabels, repeat=length):
        yield _gs.GateString(gateTuple)


def list_all_gatestrings_without_powers_and_cycles(gateLabels, maxLength):

    #Are we trying to add a germ that is a permutation of a germ we already have?  False if no, True if yes.
    def perm_check(testStr,strList): # works with python strings, so can use "in" to test for substring inclusion
        return any( [ testStr in s*2 for s in strList ] )

    #Are we trying to add a germ that is a power of a germ we already have?  False if no, True if yes.
    def pow_check(testStr,strListDict):
        L = len(testStr)
        for k in list(strListDict.keys()):
            if L % k == 0:
                rep = L // k
                if any([testStr == s*rep for s in strListDict[k] ]):
                    return True
        return False

    outputDict = {}
    for length in _np.arange(1,maxLength+1):

        permCheckedStrs = []
        for s in gen_all_gatestrings_onelen(gateLabels, length):
            pys = s.to_pythonstr(gateLabels)
            if not perm_check(pys,permCheckedStrs):#Sequence is not a cycle of anything in permCheckedStrs
                permCheckedStrs.append(pys)

        outputDict[length] = []
        for pys in permCheckedStrs:#Now check to see if any elements of tempList2 are powers of elements already in output
            if not pow_check(pys,outputDict):#Seqeunce is not a power of anything in output
                outputDict[length].append(pys)

    output = []
    for length in _np.arange(1,maxLength+1):
        output.extend( [ _gs.GateString.from_pythonstr(pys, gateLabels) for pys in outputDict[length] ] )
    return output


def list_random_gatestrings_onelen(gateLabels, length, count, seed=None):
    """
    Create a list of random gate strings of a given length.

    Parameters
    ----------
    gateLabels : tuple
        tuple of gate labels to include in gate strings.

    length : int
        the gate string length.

    count : int
        the number of random strings to create.

    seed : int, optional
        If not None, a seed for numpy's random number generator.


    Returns
    -------
    list of GateStrings
        A list of random gate strings as GateString objects.
    """
    ret = [ ]
    rndm = _rndm.RandomState(seed) # ok if seed is None
    for i in range(count): #pylint: disable=unused-variable
        r = rndm.random_sample(length) * len(gateLabels)
        ret.append( _gs.GateString( [gateLabels[int(k)] for k in r]) )
    return ret

def list_partial_strings(gateString):
    """
    List the parial strings of gateString, that is,
      the strings that are the slices gateString[0:n]
      for 0 <= l <= len(gateString).

    Parameters
    ----------
    gateString : tuple of gate labels or GateString
        The gate string to act upon.

    Returns
    -------
    list of GateString objects.
       The parial gate strings.
    """
    ret = [ ]
    for l in range(len(gateString)+1):
        ret.append( tuple(gateString[0:l]) )
    return ret

def list_lgst_gatestrings(specs, gateLabels):
    """
    List the gate strings required for runnsing LGST.

    Parameters
    ----------
    specs : 2-tuple
        A (prepSpecs,effectSpecs) tuple usually generated by calling
        build_spam_specs(...).

    gateLabels : tuple
        tuple of gate labels to estimate using LGST.

    Returns
    -------
    list of GateString objects
        The list of required gate strings, without duplicates.
    """
    rStrings, eStrings = _get_spam_strs(specs)
    singleGates = [ _gs.GateString( (gl,), "(%s)" % gl ) for gl in gateLabels ]
    ret = create_gatestring_list('eStr','prepStr','prepStr+eStr','prepStr+g+eStr',
                               eStr=eStrings, prepStr=rStrings, g=singleGates,
                               order=['g','prepStr','eStr'] ) # LEXICOGRAPHICAL VS MATRIX ORDER
    return _lt.remove_duplicates(ret)


def list_strings_lgst_can_estimate(dataset, specs):
    """
      Compute the gate strings that LGST is able to estimate
      given a set of fiducial strings or prepSpecs and effectSpecs.

      Parameters
      ----------
      dataset : DataSet
          The data used to generate the LGST estimates

      specs : 2-tuple
          A (prepSpecs,effectSpecs) tuple usually generated by calling
          build_spam_specs(...)

      Returns
      -------
      list of lists of tuples
         each list of tuples specifyies a gate string that LGST can estimate.

    """

    #Process input parameters
    prepSpecs, effectSpecs = specs

    estimatable = []
    gateStrings = list(dataset.keys())
    pre = tuple(effectSpecs[0].str);     l0 = len(pre)   #the first effectSpec string prefix
    post = tuple(prepSpecs[0].str); l1 = len(post)  #the first prepSpec string postfix

    def root_is_ok(rootStr):
        for espec in effectSpecs:
            for rhospec in prepSpecs:
                if tuple(rhospec.str) + tuple(rootStr) + tuple(espec.str) not in gateStrings: # LEXICOGRAPHICAL VS MATRIX ORDER
                    return False
        return True

    #check if string has first fiducial at beginning & end, and if so
    # strip that first fiducial off, leaving a 'root' string that we can test
    for s in gateStrings:
        if s[0:l0] == pre and s[len(s)-l1:] == post:
            root = s[l0:len(s)-l1]
            if root_is_ok( root ):
                estimatable.append( root )

    return gatestring_list(estimatable)



def gatestring_list( listOfGateLabelTuplesOrStrings ):
    """
    Converts a list of gate label tuples or strings to
     a list of GateString objects.

    Parameters
    ----------
    listOfGateLabelTuplesOrStrings : list
        List which may contain a mix of GateString objects, tuples of gate
        labels, and strings in standard-text-format.

    Returns
    -------
    list of GateString objects
        Each item of listOfGateLabelTuplesOrStrings converted to a GateString.
    """
    ret = []
    for x in listOfGateLabelTuplesOrStrings:
        if isinstance(x,_gs.GateString):
            ret.append(x)
        elif isinstance(x,tuple) or isinstance(x,list):
            ret.append( _gs.GateString(x) )
        elif isinstance(x,str):
            ret.append( _gs.GateString(None, x) )
        else:
            raise ValueError("Cannot convert type %s into a GateString" % str(type(x)))
    return ret


def translate_gatestring_list(gatestringList, aliasDict):
    """
    Creates a new list of GateString objects from an existing one by replacing
    gate labels in `gatestringList` by (possibly multiple) new labels according
    to `aliasDict`.

    Parameters
    ----------
    gatestringList : list of GateStrings
        The list of gate strings to use as the base for find & replace
        operations.

    aliasDict : dict
        A dictionary whose keys are single gate labels and whose values are 
        lists or tuples of the new gate labels that should replace that key.

    Returns
    -------
    list of GateStrings
    """
    new_gatestrings = [ _gs.GateString(tuple(_itertools.chain(
                *[aliasDict.get(lbl,lbl) for lbl in gs])))
                        for gs in gatestringList ]
    return new_gatestrings


def compose_alias_dicts(aliasDict1, aliasDict2):
    """
    Composes two alias dicts.
    
    Assumes `aliasDict1` maps "A" labels to "B" labels and `aliasDict2` maps
    "B" labels to "C" labels.  The returned dictionary then maps "A" labels
    directly to "C" labels, and satisfies:

    `returned[A_label] = aliasDict2[ aliasDict1[ A_label ] ]`
    
    Parameters
    ----------
    aliasDict1, aliasDict2 : dict
        The two dictionaries to compose.

    Returns
    -------
    dict
    """
    ret = {}
    for A,Bs in aliasDict1.items():
        ret[A] = list(_itertools.chain(*[aliasDict2[B] for B in Bs]))
    return ret



#Unneeded
#def list_periodic_gatestrings(gateLabels, max_period, minlength, maxlength, left_bookends=[()], right_bookends=[()]):
#    ret = [ ]
#    for lb in left_bookends:
#        for rb in right_bookends:
#            for l in range(minlength, maxlength+1):
#                pdic = list_periodic_gatestrings_onelen(gateLabels, max_period, l)
#                ret += [ tuple(lb) + tuple(p) + tuple(rb) for p in pdic ]
#    return _lt.remove_duplicates(ret)
#
#def list_periodic_gatestrings_onelen(gateLabels, max_period, length):
#    ret = [ ]
#    if max_period >= 0: ret.append( [] )
#    for period_length in range(1,min(max_period,length)+1):
#        nPeriods = _np.ceil(length / float(period_length))
#        for period in list_all_gatestrings_onelen(gateLabels, period_length):
#            for k in range(1,period_length):
#                if period_length % k > 0: continue
#                if period in list_periodic_gatestrings_onelen(gateLabels, k, period_length):
#                    break # period is itself periodic with period k < len(period), so don't use it as the string it generates has already been found
#            else:
#                s = period * nPeriods
#                ret.append( tuple(s[0:length]) )
#    return _lt.remove_duplicates(ret)
#
#
#
#def list_exponentiated_germ_gatestrings(germs, exponents, ends=None, left_ends=None, right_ends=None):
#    if ends is not None:
#        if not left_ends and not right_ends:
#            left_ends = right_ends = ends
#        else: raise ValueError("Conflicting arguments to list_exponentiated_germ_gatestrings - specify either" + \
#                                   "'ends' or 'left_ends' and/or 'right_ends', not both")
#    if left_ends is None: left_ends = [()]
#    elif () not in left_ends: left_ends = [()] + left_ends
#    if right_ends is None: right_ends = [()]
#    elif () not in right_ends: right_ends = [()] + right_ends
#
#    ret = create_gatestring_list("lb+germ*exp+rb", germ=germs, exp=exponents,lb=left_ends, rb=right_ends,
#                               order=['germ','exp','lb','rb'] )
#    return ret
#
#
#def read_gatestring_list(filename,**kwargs):
#    emptyCode = kwargs.get("empty_code","") # code for the empty string
#    mode = kwargs.get("mode", "comma-delim") #or const-length
#    L = kwargs.get("length", 0)
#    if mode == "const-length": assert(L > 0)
#
#    ret = [ ]
#    for line in open(filename):
#        if line[0][0]=='#': continue
#        splitline = line.split()
#        if len(splitline) == 0: continue
#        charStr = splitline[0]
#        if charStr == emptyCode:
#            gateString = ()
#        elif mode == "comma-delim":
#            gateString = tuple( charStr.split(",") )
#        elif mode == "const-length":
#            gateString = tuple( [ charStr[i:i+L] for i in range(0,len(charStr),L) ] )
#        else:
#            raise ValueError("Invalid mode passed to read_gatestring_list: %s" % mode)
#
#        ret.append(gateString)
#
#    return ret



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Functions for generating bootstrapped error bars """
import numpy as _np
import matplotlib as _mpl
from .longsequence import do_long_sequence_gst as _do_long_sequence_gst

from .. import objects as _obj
from .. import algorithms as _alg
from .. import tools as _tools

def make_bootstrap_dataset(inputDataSet,generationMethod,inputGateSet=None,
                           seed=None,spamLabels=None,verbosity=1):
    """
    Creates a DataSet used for generating bootstrapped error bars.

    Parameters
    ----------
    inputDataSet : DataSet
       The data set to use for generating the "bootstrapped" data set.

    generationMethod : { 'nonparametric', 'parametric' }
      The type of dataset to generate.  'parametric' generates a DataSet
      with the same gate strings and sample counts as inputDataSet but
      using the probabilities in inputGateSet (which must be provided).
      'nonparametric' generates a DataSet with the same gate strings
      and sample counts as inputDataSet using the count frequencies of
      inputDataSet as probabilities.

    inputGateSet : GateSet, optional
       The gate set used to compute the probabilities for gate strings when
       generationMethod is set to 'parametric'.  If 'nonparametric' is selected,
       this argument must be set to None (the default).

    seed : int, optional
       A seed value for numpy's random number generator.

    spamLabels : list, optional
       The list of SPAM labels to include in the output dataset.  If None
       are specified, defaults to the spam labels of inputDataSet.

    verbosity : int, optional
       How verbose the function output is.  If 0, then printing is suppressed.
       If 1 (or greater), then printing is not suppressed.

    Returns
    -------
    DataSet
    """
    if generationMethod not in ['nonparametric', 'parametric']:
        raise ValueError("generationMethod must be 'parametric' or 'nonparametric'!")
    if spamLabels is None:
        spamLabels = inputDataSet.get_spam_labels()

    rndm = _np.random.RandomState(seed)
    if inputGateSet is None:
        if generationMethod == 'nonparametric':
            print("Generating non-parametric dataset.")
        elif generationMethod == 'parametric':
            raise ValueError("For 'parmametric', must specify inputGateSet")
    else:
        if generationMethod == 'parametric':
            print("Generating parametric dataset.")
        elif generationMethod == 'nonparametric':
            raise ValueError("For 'nonparametric', inputGateSet must be None")
        possibleSpamLabels = inputGateSet.get_spam_labels()
        assert( all([sl in possibleSpamLabels for sl in spamLabels]) )

    possibleSpamLabels = inputDataSet.get_spam_labels()
    assert( all([sl in possibleSpamLabels for sl in spamLabels]) )

    #create new dataset
    simDS = _obj.DataSet(spamLabels=spamLabels, 
                         collisionAction=inputDataSet.collisionAction)
    gatestring_list = list(inputDataSet.keys())
    for s in gatestring_list:
        nSamples = inputDataSet[s].total()
        if generationMethod == 'parametric':
            ps = inputGateSet.probs(s)
        elif generationMethod == 'nonparametric':
            ps = { sl: inputDataSet[s].fraction(sl) for sl in spamLabels }
        pList = _np.array([_np.clip(ps[spamLabel],0,1) for spamLabel in spamLabels])
          #Truncate before normalization; bad extremal values shouldn't
          # screw up not-bad values, yes?
        pList = pList / sum(pList)
        countsArray = rndm.multinomial(nSamples, pList, 1)
        counts = { sl: countsArray[0,i] for i,sl in enumerate(spamLabels) }
        simDS.add_count_dict(s, counts)
    simDS.done_adding_data()
    return simDS

def make_bootstrap_gatesets(numGateSets, inputDataSet, generationMethod,
                            fiducialPrep, fiducialMeasure, germs, maxLengths,
                            inputGateSet=None, targetGateSet=None, startSeed=0,
                            spamLabels=None, lsgstLists=None,
                            returnData=False, verbosity=2):
    """
    Creates a series of "bootstrapped" GateSets form a single DataSet (and
    possibly GateSet) used for generating bootstrapped error bars.  The
    resulting GateSets are obtained by performing MLGST on datasets generated
    by repeatedly calling make_bootstrap_dataset with consecutive integer seed
    values.

    Parameters
    ----------
    numGateSets : int
       The number of gate sets to create.

    inputDataSet : DataSet
       The data set to use for generating the "bootstrapped" data set.

    generationMethod : { 'nonparametric', 'parametric' }
      The type of datasets to generate.  'parametric' generates DataSets
      with the same gate strings and sample counts as inputDataSet but
      using the probabilities in inputGateSet (which must be provided).
      'nonparametric' generates DataSets with the same gate strings
      and sample counts as inputDataSet using the count frequencies of
      inputDataSet as probabilities.

    fiducialPrep : list of GateStrings
        The state preparation fiducial gate strings used by MLGST.

    fiducialMeasure : list of GateStrings
        The measurement fiducial gate strings used by MLGST.

    germs : list of GateStrings
        The germ gate strings used by MLGST.

    maxLengths : list of ints
        List of integers, one per MLGST iteration, which set truncation lengths
        for repeated germ strings.  The list of gate strings for the i-th LSGST
        iteration includes the repeated germs truncated to the L-values *up to*
        and including the i-th one.

    inputGateSet : GateSet, optional
       The gate set used to compute the probabilities for gate strings when
       generationMethod is set to 'parametric'.  If 'nonparametric' is selected,
       this argument must be set to None (the default).

    targetGateSet : GateSet, optional
       Mandatory gate set to use for as the target gate set for MLGST when
       generationMethod is set to 'nonparametric'.  When 'parametric'
       is selected, inputGateSet is used as the target.

    startSeed : int, optional
       The initial seed value for numpy's random number generator when
       generating data sets.  For each succesive dataset (and gateset)
       that are generated, the seed is incremented by one.

    spamLabels : list, optional
       The list of SPAM labels to include in the output dataset.  If None
       are specified, defaults to the spam labels of inputDataSet.

    lsgstLists : list of gate string lists, optional
        Provides explicit list of gate string lists to be used in analysis;
        to be given if the dataset uses "incomplete" or "reduced" sets of
        gate string.  Default is None.

    returnData : bool
        Whether generated data sets should be returned in addition to
        gate sets.

    verbosity : int
        Level of detail printed to stdout.

    Returns
    -------
    gatesets : list
       The list of generated GateSet objects.

    datasets : list
       The list of generated DataSet objects, only returned when
       returnData == True.
    """

    if maxLengths == None:
        print("No maxLengths value specified; using [0,1,24,...,1024]")
        maxLengths = [0]+[2**k for k in range(10)]

    if (inputGateSet is None and targetGateSet is None):
        raise ValueError("Must supply either inputGateSet or targetGateSet!")
    if (inputGateSet is not None and targetGateSet is not None):
        raise ValueError("Cannot supply both inputGateSet and targetGateSet!")

    if generationMethod == 'parametric':
        targetGateSet = inputGateSet

    datasetList = []
    print("Creating DataSets: ")
    for run in range(numGateSets):
        print("%d " % run, end='')
        datasetList.append(
            make_bootstrap_dataset(inputDataSet,generationMethod,
                                   inputGateSet, startSeed+run,
                                   spamLabels)
            )

    gatesetList = []
    print("Creating GateSets: ")
    for run in range(numGateSets):
        print("Running MLGST Iteration %d " % run)
        results = _do_long_sequence_gst(
            datasetList[run], targetGateSet, fiducialPrep, fiducialMeasure,
            germs, maxLengths, lsgstLists=lsgstLists, verbosity=verbosity)
        gatesetList.append(results.gatesets['final estimate'])

    if not returnData:
        return gatesetList
    else:
        return gatesetList, datasetList


def gauge_optimize_gs_list(gsList, targetGateset,
                           gateMetric = 'frobenius', spamMetric = 'frobenius',
                           plot=True):
    """
    Optimizes the "spam weight" parameter used in gauge optimization by
    attempting spam a range of spam weights and taking the one the minimizes
    the average spam error multiplied by the average gate error (with respect
    to a target gate set).

    Parameters
    ----------
    gsList : list
       The list of GateSet objects to gauge optimize (simultaneously).

    targetGateset : GateSet
       The gateset to compare the gauge-optimized gates with, and also
       to gauge-optimize them to.

    gateMetric : { "frobenius", "fidelity", "tracedist" }, optional
       The metric used within the gauge optimization to determing error
       in the gates.

    spamMetric : { "frobenius", "fidelity", "tracedist" }, optional
       The metric used within the gauge optimization to determing error
       in the state preparation and measurement.

    plot : bool, optional
       Whether to create a plot of the gateset-target discrepancy
       as a function of spam weight (figure displayed interactively).

    Returns
    -------
    list
       The list of GateSets gauge-optimized using the best spamWeight.
    """

    listOfBootStrapEstsNoOpt = list(gsList)
    numResamples = len(listOfBootStrapEstsNoOpt)
    ddof = 1
    SPAMMin = []
    SPAMMax = []
    SPAMMean = []

    gateMin = []
    gateMax = []
    gateMean = []
    for spWind, spW in enumerate(_np.logspace(-4,0,13)): #try spam weights
        print("Spam weight %s" % spWind)
        listOfBootStrapEstsNoOptG0toTargetVarSpam = []
        for gs in listOfBootStrapEstsNoOpt:
            listOfBootStrapEstsNoOptG0toTargetVarSpam.append(
                _alg.gaugeopt_to_target(gs,targetGateset, 
                                        itemWeights={'spam': spW },
                                        gatesMetric=gateMetric,
                                        spamMetric=spamMetric))

        GateSetGOtoTargetVarSpamVecArray = _np.zeros([numResamples],
                                                     dtype='object')
        for i in range(numResamples):
            GateSetGOtoTargetVarSpamVecArray[i] = \
                listOfBootStrapEstsNoOptG0toTargetVarSpam[i].to_vector()

        gsStdevVec = _np.std(GateSetGOtoTargetVarSpamVecArray,ddof=ddof)
        gsStdevVecSPAM = gsStdevVec[:8]
        gsStdevVecGates = gsStdevVec[8:]

        SPAMMin.append(_np.min(gsStdevVecSPAM))
        SPAMMax.append(_np.max(gsStdevVecSPAM))
        SPAMMean.append(_np.mean(gsStdevVecSPAM))

        gateMin.append(_np.min(gsStdevVecGates))
        gateMax.append(_np.max(gsStdevVecGates))
        gateMean.append(_np.mean(gsStdevVecGates))

    if plot:
        _mpl.pyplot.loglog(_np.logspace(-4,0,13),SPAMMean,'b-o')
        _mpl.pyplot.loglog(_np.logspace(-4,0,13),SPAMMin,'b--+')
        _mpl.pyplot.loglog(_np.logspace(-4,0,13),SPAMMax,'b--x')

        _mpl.pyplot.loglog(_np.logspace(-4,0,13),gateMean,'r-o')
        _mpl.pyplot.loglog(_np.logspace(-4,0,13),gateMin,'r--+')
        _mpl.pyplot.loglog(_np.logspace(-4,0,13),gateMax,'r--x')

        _mpl.pyplot.xlabel('SPAM weight in gauge optimization')
        _mpl.pyplot.ylabel('Per element error bar size')
        _mpl.pyplot.title('Per element error bar size vs. ${\\tt spamWeight}$')
        _mpl.pyplot.xlim(1e-4,1)
        _mpl.pyplot.legend(['SPAM-mean','SPAM-min','SPAM-max',
                            'gates-mean','gates-min','gates-max'],
                           bbox_to_anchor=(1.4, 1.))

    # gateTimesSPAMMean = _np.array(SPAMMean) * _np.array(gateMean)

    bestSPAMWeight = _np.logspace(-4,0,13)[ _np.argmin(
            _np.array(SPAMMean)*_np.array(gateMean)) ]
    print("Best SPAM weight is %s" % bestSPAMWeight)

    listOfBootStrapEstsG0toTargetSmallSpam = []
    for gs in listOfBootStrapEstsNoOpt:
        listOfBootStrapEstsG0toTargetSmallSpam.append(
            _alg.gaugeopt_to_target(gs,targetGateset,
                                    itemWeights={'spam': bestSPAMWeight},
                                    gatesMetric=gateMetric,
                                    spamMetric=spamMetric))

    return listOfBootStrapEstsG0toTargetSmallSpam




################################################################################
#TODO: need to add docstrings and perhaps relocate the utility functions below #
################################################################################

#For metrics that evaluate gateset with single scalar:
def gs_stdev(gsFunc, gsEnsemble, ddof=1, **kwargs):
    return _np.std([gsFunc(gs, **kwargs) for gs in gsEnsemble],ddof=ddof)

def gs_mean(gsFunc, gsEnsemble, axis = 0,**kwargs):
    return _np.mean([gsFunc(gs, **kwargs) for gs in gsEnsemble])

#For metrics that evaluate gateset with scalar for each gate
def gs_stdev1(gsFunc, gsEnsemble, ddof=1,axis=0, **kwargs):
    return _np.std([gsFunc(gs, **kwargs) for gs in gsEnsemble],axis=axis,ddof=ddof)

def gs_mean1(gsFunc, gsEnsemble, axis = 0,**kwargs):
    return _np.mean([gsFunc(gs,**kwargs) for gs in gsEnsemble],axis=axis)

def to_vector(gs):
    return gs.to_vector()

def to_mean_gateset(gsList,target_gs):
    numResamples = len(gsList)
    gsVecArray = _np.zeros([numResamples],dtype='object')
    for i in range(numResamples):
        gsVecArray[i] = gsList[i].to_vector()
    output_gs = target_gs.copy()
    output_gs.from_vector(_np.mean(gsVecArray))
    return output_gs

def to_std_gateset(gsList,target_gs,ddof=1):
    numResamples = len(gsList)
    gsVecArray = _np.zeros([numResamples],dtype='object')
    for i in range(numResamples):
        gsVecArray[i] = gsList[i].to_vector()
    output_gs = target_gs.copy()
    output_gs.from_vector(_np.std(gsVecArray,ddof=ddof))
    return output_gs

def to_rms_gateset(gsList,target_gs):
    numResamples = len(gsList)
    gsVecArray = _np.zeros([numResamples],dtype='object')
    for i in range(numResamples):
        gsVecArray[i] = _np.sqrt(gsList[i].to_vector()**2)
    output_gs = target_gs.copy()
    output_gs.from_vector(_np.mean(gsVecArray))
    return output_gs

def gateset_jtracedist(gs,gs_target,mxBasis="gm"):
    output = _np.zeros(3,dtype=float)
    for i, gate in enumerate(gs_target.gates.keys()):
        output[i] = _tools.jtracedist(gs.gates[gate],gs_target.gates[gate],mxBasis=mxBasis)
#    print output
    return output

def gateset_process_fidelity(gs,gs_target):
    output = _np.zeros(3,dtype=float)
    for i, gate in enumerate(gs_target.gates.keys()):
        output[i] = _tools.process_fidelity(gs.gates[gate],gs_target.gates[gate])
    return output

def gateset_decomp_angle(gs):
    output = _np.zeros(3,dtype=float)
    for i, gate in enumerate(gs.gates.keys()):
        output[i] = _tools.decompose_gate_matrix(gs.gates[gate]).get('pi rotations',0)
    return output

def gateset_decomp_decay_diag(gs):
    output = _np.zeros(3,dtype=float)
    for i, gate in enumerate(gs.gates.keys()):
        output[i] = _tools.decompose_gate_matrix(gs.gates[gate]).get('decay of diagonal rotation terms',0)
    return output

def gateset_decomp_decay_offdiag(gs):
    output = _np.zeros(3,dtype=float)
    for i, gate in enumerate(gs.gates.keys()):
        output[i] = _tools.decompose_gate_matrix(gs.gates[gate]).get('decay of off diagonal rotation terms',0)
    return output

#def gateset_fidelity(gs,gs_target,mxBasis="gm"):
#    output = _np.zeros(3,dtype=float)
#    for i, gate in enumerate(gs_target.gates.keys()):
#        output[i] = _tools.fidelity(gs.gates[gate],gs_target.gates[gate])
#    return output

def gateset_diamonddist(gs,gs_target,mxBasis="gm"):
    output = _np.zeros(3,dtype=float)
    for i, gate in enumerate(gs_target.gates.keys()):
        output[i] = _tools.diamonddist(gs.gates[gate],gs_target.gates[gate],mxBasis=mxBasis)
    return output

def spamrameter(gs):
    firstRho = list(gs.preps.keys())[0]
    firstE = list(gs.effects.keys())[0]
    return _np.dot(gs.preps[firstRho].T,gs.effects[firstE])[0,0]



from __future__ import print_function
import os, site

myPackagePath = os.path.join( os.path.dirname(os.path.abspath(__file__)), "packages")

pathspec = r"""
# Generated by GST's installer (install_GST.py)
# In the lines below, list the paths where Python should look for
# GateSetTomography-supplied modules, one directory per line.
#
# If a directory does not exist when Python is started, it will be ignored.
%s
""" % myPackagePath

print("Adding path:",myPackagePath)

usp = site.getusersitepackages()
if not os.path.exists(usp):
    os.makedirs(usp)
uspfile = os.path.join(usp, 'GST.pth')
open(uspfile, 'w').write(pathspec)
print('Wrote to ' + uspfile)

print("GST package installed successfully!")



from __future__ import division, print_function, absolute_import, unicode_literals

#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************

""" Functions for generating report tables in different formats """

from .html  import html,  html_value
from .latex import latex, latex_value
from .ppt   import ppt,   ppt_value

from inspect import getargspec as _getargspec

import cgi     as _cgi
import numpy   as _np
import numbers as _numbers
import re      as _re
import os      as _os

def _give_specs(formatter, specs):
    '''
    Pass parameters down to a formatter

    Parameters
    --------
    formatter : callable, takes arguments

    specs : dictionary of argnames : values

    Returns
    ------
    None

    Raises
    ------
    ValueError : If a needed spec is not supplied.
    '''
    # If the formatter requires a setting to do its job, give the setting
    if hasattr(formatter, 'specs'):
        for spec in formatter.specs:
            if spec not in specs or specs[spec] is None:
                '''
                This should make the ValueError thrown by
                  _ParameterizedFormatter redundant
                This also means that even though specs will be set after
                the first call to table.render(),
                they will need to be provided again in subsequent calls
                '''
                raise ValueError(
                        ('The spec %s was not supplied to ' % spec) +
                        ('FormatSet, but is needed by an active formatter'))
            formatter.specs[spec] = specs[spec]

class FormatSet():
    '''
    Attributes
    ---------
    formatDict: Static dictionary containing small formatter dictionaries
                Ex: { 'Rho' :  { 'html' : ... , 'text' : ... }, ... }
                (created below)

    Methods
    -------

    __init__: (specs): Specs is a dictionary of the form { 'setting'(kwarg) : value }
                       Ex: { 'precision' : 6, 'polarprecision' : 3, 'sciprecision' : 0 }
                       given to _ParameterizedFormatters that need them

    formatList : Given a list of items and formatters and a target format, returns formatted items
    '''
    formatDict = {}

    def __init__(self, specs):
        self.specs = specs

    def formatList(self, items, formatterNames, fmt):
        assert(len(items) == len(formatterNames))
        formatted_items = []

        for item, formatterName in zip(items, formatterNames):
            if formatterName is not None:
                formatter = FormatSet.formatDict[formatterName]
                _give_specs(formatter[fmt], self.specs) # Parameters aren't sent until render call
                # Format the item once the formatter has been completely built
                formatted_item = formatter[fmt](item)
                if formatted_item is None:
                    raise ValueError("Formatter " + str(type(formatter[fmt]))
                                     + " returned None for item = " + str(item))
                if isinstance(formatted_item, list): #formatters can create multiple table cells by returning *lists* 
                    formatted_items.extend( formatted_item )
                else:
                    formatted_items.append( formatted_item )
            else:
                if item is None:
                    raise ValueError("Unformatted None in formatList")
                formatted_items.append( item )

        return formatted_items

class _Formatter(object):
    '''
    Callable class that can replace a formatter function.

    Only defines __init__ and __call__ methods
    '''

    def __init__(self, stringreplacers=None, regexreplace=None,
                       formatstring='%s', stringreturn=None):
        '''
        Parameters
        ----------
        stringreplacers : tuples of the form (pattern, replacement) (optional)
                       (replacement is a normal string)
                     Ex : [('rho', '&rho;')]
        regexreplace  : A tuple of the form (regex,   replacement) (optional)
                       (replacement is formattable string,
                          gets formatted with grouped result of regex matching on label)
                     Ex : ('.*?([0-9]+)$', '_{%s}')

        formatstring : string (optional) Outer formatting for after both replacements have been made

        stringreturn : tuple (string, string) Replaces first string with second and
                         returns early if the first string exists,
                         otherwise does nothing
        '''
        self.stringreplacers = stringreplacers
        self.regexreplace    = regexreplace
        self.formatstring    = formatstring
        self.stringreturn    = stringreturn

    def __call__(self, label):
        '''
        Formatting function template

        Parameters
        --------
        label : string, the label to be formatted!

        Returns
        --------
        formatted label : string
        '''
        # Exit early if string matches stringreturn
        if self.stringreturn is not None and self.stringreturn[0] == label:
            return self.stringreturn[1]
            #Changed by EGN: no need to format string here, but do need to
            # check for equality above

        # Below is the standard formatter case:
        # Replace all occurances of certain substrings
        if self.stringreplacers is not None:
            for stringreplace in self.stringreplacers:
                label = label.replace(stringreplace[0], stringreplace[1])
        # And then replace all occurances of certain regexes
        if self.regexreplace is not None:
            result = _re.match(self.regexreplace[0], label)
            if result is not None:
                grouped = result.group(1)
                label   = label[0:-len(grouped)] + (self.regexreplace[1] % grouped)
        # Additional formatting, ex $%s$ or <i>%s</i>
        return self.formatstring % label


class _TupleFormatter(object):
    '''
    Callable class that can replace a formatter function, similar to
    _Formatter, but expects a tuple as input instead of a single string.

    Only defines __init__ and __call__ methods
    '''

    def __init__(self, label_formatter=None, formatstring='{l0}'):
        '''
        Parameters
        ----------
        label_formatter : callable or None
            Another formatter that is used to format the "label",
            defined to be the first element of the tuple this 
            formatter is called with.

        formatstring : string (optional)
            Outer formatting for after label_formatter has been applied.
        '''
        self.formatstring    = formatstring
        self.label_formatter = label_formatter

    def __call__(self, label_tuple):
        '''
        Formatting function template

        Parameters
        --------
        label_tuple : tuple
            The label, followed by other paramters, to be formatted.

        Returns
        --------
        formatted label : string
        '''
        label = label_tuple[0] #process first element of tuple as _Formatter

        if self.label_formatter is not None:
            label = self.label_formatter(label)

        # Formatting according to format string
        format_dict = { 'l0': label }
        format_dict.update( { 'l%d' % i: label_tuple[i] 
                              for i in range(1,len(label_tuple)) })
        return self.formatstring.format(**format_dict)

def _no_format(label):
    return label

# Helper function to _ParameterizedFormatter
def _has_argname(argname, function):
    return argname in _getargspec(function).args

# Gives arguments to formatters
class _ParameterizedFormatter(object):
    '''
    Class that will pass down specs (arguments) to functions that need them

    For example, a precision-parameterized latex formatter without the help of the _PrecisionFormatter might look like this:
    formatter = _ParameterizedFormatter(latex, ['precision', 'polarprecision', 'sciprecision'])
    Which, when used with a FormatSet, would have arguments to table.render() passed down to the latex() function
    '''
    def __init__(self, custom, neededSpecs, defaults={}, formatstring='%s'):
        self.custom       = custom
        self.specs        = { neededSpec : None for neededSpec in neededSpecs }
        self.defaults     = defaults
        self.formatstring = formatstring

    def __call__(self, label):
        # If the formatter is being called, we know that the needed specs have successfully been supplied by FormatSet
        self.defaults.update(self.specs)
        # Supply arguments to the custom formatter (if it needs them)
        for argname in self.defaults:
            if not callable(self.custom): # If some keyword arguments were supplied already
                if _has_argname(argname, self.custom[0]):             # 'if it needs them'
                    # update the argument in custom's existing keyword dictionary
                    self.custom[1][argname] = self.defaults[argname]
            else:
                if _has_argname(argname, self.custom): # If custom is a lone callable (not a tuple)
                # Create keyword dictionary for custom, modifiying it to be a tuple
                #   (function, kwargs)
                    self.custom = (self.custom, {argname : self.defaults[argname]})
        return self.formatstring % self.custom[0](label, **self.custom[1])

# Gives precision arguments to formatters
class _PrecisionFormatter(_ParameterizedFormatter):
    '''Helper class for Precision Formatting
    Takes a custom function and a dictionary of keyword arguments:
    So, something like _PrecisionFormatter(latex) would pass precision arguments to
      the latex formatter function during table.render() calls
    '''
    def __init__(self, custom, defaults={}, formatstring='%s'):
        super(_PrecisionFormatter, self).__init__(custom, ['precision', 'polarprecision','sciprecision'],
                                                 defaults, formatstring)


# Formatter class that requires a scratchDirectory from an instance of FormatSet for saving figures to
class _FigureFormatter(_ParameterizedFormatter):
    '''
    Helper class that utilizes a scratchDir variable to render figures
    '''
    def __init__(self, extension='.png', formatstring='%s%s%s%s'):
        '''
        Parameters
        ---------
        extension : string, optional. extension of the figure's image
        formatstring : string, optional. Normally formatted with W, H, scratchDir, filename
        '''
        super(_FigureFormatter, self).__init__(_no_format, ['scratchDir'])
        self.extension    = extension
        self.formatstring = formatstring

    # Override call method of Parameterized formatter
    def __call__(self, figInfo):
        fig, name, W, H = figInfo
        scratchDir = self.specs['scratchDir']
        if len(scratchDir) > 0: #empty scratchDir signals not to output figure
            fig.save_to(_os.path.join(scratchDir, name + self.extension))
        return self.formatstring % (W, H, scratchDir,
                                    name + self.extension)

# Takes two formatters (a and b), and determines which to use based on a predicate
# (Used in building formatter templates)
class _BranchingFormatter(object):
    def __init__(self, predicate, a, b):
        self.predicate = predicate
        self.a = a
        self.b = b

        # So that a branching formatter can hold parameterized formatters
        self.specs = {}
        if hasattr(a, 'specs'):
            self.specs.update(a.specs)
        if hasattr(b, 'specs'):
            self.specs.update(b.specs)

    def __call__(self, label):
        if self.predicate(label):
            _give_specs(self.a, self.specs)
            return self.a(label)
        else:
            _give_specs(self.a, self.specs)
            return self.b(label)

##############################################################################
#                          Formatting functions                              #
##############################################################################

# 'rho' (state prep) formatting
# Replace rho with &rho;
# Numbers following 'rho' -> subscripts
FormatSet.formatDict['Rho'] = {
    'html'  : _Formatter(stringreplacers=[('rho', '&rho;')],
                         regexreplace=('.*?([0-9]+)$', '<sub>%s</sub>')),
    'latex' : _Formatter(stringreplacers=[('rho', '\\rho')],
                         regexreplace=('.*?([0-9]+)$', '_{%s}'), formatstring='$%s$'),
    'text'  : _no_format,
    'ppt'   : _no_format}

# 'E' (POVM) effect formatting
FormatSet.formatDict['Effect'] = {
    # If label == 'remainder', return E sub C
    # Otherwise, match regex and replace with subscript
    'html'  : _Formatter(stringreturn=('remainder', 'E<sub>C</sub>'),
                         regexreplace=('.*?([0-9]+)$', '<sub>%s</sub>')),
    'latex' : _Formatter(stringreturn=('remainder', '$E_C$'),
                         regexreplace=('.*?([0-9]+)$', '_{%s}'), formatstring='$%s$'),
    'text'  : _no_format,
    'ppt'   : _no_format}

# Normal replacements
FormatSet.formatDict['Normal'] = {
    'html'  : _PrecisionFormatter(html),
    'latex' : _PrecisionFormatter(latex),
    'text'  : _no_format,
    'ppt'   : _PrecisionFormatter(ppt) }

# 'normal' formatting but round to 2 decimal places regardless of what is passed in to table.render()
FormatSet.formatDict['Rounded'] = {
    'html'  : _ParameterizedFormatter(html_value,  ['polarprecision'], {'precision' : 2, 'sciprecision': 0}),
    'latex' : _ParameterizedFormatter(latex_value, ['polarprecision'], {'precision' : 2, 'sciprecision': 0}),
    'text'  : _no_format,
    'ppt'   : _ParameterizedFormatter(ppt_value,   ['polarprecision'], {'precision' : 2, 'sciprecision': 0})}

# Similar to the above two formatdicts,
# but recieves precision during table.render(), which is sent as kwarg to html_value, for example
FormatSet.formatDict['Precision'] = {
    'html'  : _PrecisionFormatter(html_value),
    'latex' : _PrecisionFormatter(latex_value),
    'text'  : _no_format,
    'ppt'   : _PrecisionFormatter(ppt_value)}

# 'small' formating - make text smaller
FormatSet.formatDict['Small'] = {
    'html'  : _PrecisionFormatter(html),
    'latex' : _PrecisionFormatter(latex, formatstring='\\small%s'),
    'text'  : _no_format,
    'ppt'   : _PrecisionFormatter(ppt)}

# 'small' formating - make text smaller
FormatSet.formatDict['Verbatim'] = {
    'html'  : _PrecisionFormatter(html),
    'latex' : _Formatter(formatstring='\\spverb!%s!'),
    'text'  : _no_format,
    'ppt'   : _PrecisionFormatter(ppt)}

#############################################
# Helper functions for formatting pi-labels #
#############################################

def _pi_template(b):
    # Pi Formatting shares a common predicate and first branch condition
    return _BranchingFormatter(lambda label : str(label) == '--' or str(label) == '',
                              _no_format, b)

# Requires an additional predicate
def _pi_text(label):
    if label == '--' or label == '' or not isinstance(label, _numbers.Number):
        return label
    else:
        return label * _np.pi

# Pi formatters
FormatSet.formatDict['Pi'] = {
    'html'  : _pi_template(_PrecisionFormatter(html,  formatstring='%s&pi;')),
    'latex' : _pi_template(_PrecisionFormatter(latex, formatstring='%s$\\pi$')),
    'text'  : _pi_text,
    'ppt'   : _pi_template(_PrecisionFormatter(ppt,   formatstring='%spi'))}

# Bracket Formatters
FormatSet.formatDict['Brackets'] = {
    'html'  : _PrecisionFormatter(html,  defaults={'brackets' : True}),
    'latex' : _PrecisionFormatter(latex, defaults={'brackets' : True}),
    'text'  : _no_format,
    'ppt'   : _PrecisionFormatter(ppt,   defaults={'brackets' : True})}

##################################################################################
# 'conversion' formatting: catch all for find/replacing specially formatted text #
##################################################################################

# These two formatters are more complex, justifying individual functions:

def _fmtCnv_html(x):
    x = x.replace("\\", "&#92"); #backslash
    x = x.replace("|"," ") #remove pipes=>newlines, since html wraps table text automatically
    x = x.replace("<STAR>","REPLACEWITHSTARCODE") #b/c cgi.escape would mangle <STAR> marker
    x = _cgi.escape(x).encode("ascii","xmlcharrefreplace")
    x = x.replace(b"REPLACEWITHSTARCODE", b'&#9733;') #replace new marker with HTML code
    return x

def _fmtCnv_latex(x):
    x = x.replace("\\", "\\textbackslash")
    x = x.replace('%','\\%')
    x = x.replace('#','\\#')
    x = x.replace("half-width", "$\\nicefrac{1}{2}$-width")
    x = x.replace("1/2", "$\\nicefrac{1}{2}$")
    x = x.replace("Diamond","$\\Diamond$")
    x = x.replace("Check","\\checkmark")
    if "<STAR>" in x: #assume <STAR> never has $ around it already
        x = "$" + x.replace("<STAR>","\\bigstar") + "$"
    if "|" in x:
        return '\\begin{tabular}{c}' + '\\\\'.join(x.split("|")) + '\\end{tabular}'
    else:
        return x

FormatSet.formatDict['Conversion'] = {
    'html'  : _fmtCnv_html,
    'latex' : _fmtCnv_latex,
    'text'  : _Formatter(stringreplacers=[('<STAR>', '*'), ('|', ' ')]),
    'ppt'   : _Formatter(stringreplacers=[('<STAR>', '*'), ('|', '\n')])}

_eb_exists = lambda t : t[1] is not None

class _EBFormatter(object):
    def __init__(self, f, formatstringA='%s +/- %s', formatstringB='%s'):
        self.f = f
        if hasattr(f, 'specs'):
            self.specs = f.specs
        self.formatstringA = formatstringA
        self.formatstringB = formatstringB

    def __call__(self, t):
        if hasattr(self.f, 'specs'):
            _give_specs(self.f, self.specs)
        if _eb_exists(t):
            return self.formatstringA % (self.f(t[0]), self.f(t[1]))
        else:
            return self.formatstringB % self.f(t[0])

_EB_html  = _EBFormatter(_PrecisionFormatter(html))
_EB_latex = _EBFormatter(_PrecisionFormatter(latex_value),
                       '$ \\begin{array}{c} %s \\\\ \pm %s \\end{array} $')
_EB_text  = lambda t : {'value' : t[0], 'errbar' : t[1]}
_EB_ppt   = _EBFormatter(_PrecisionFormatter(ppt))

FormatSet.formatDict['ErrorBars'] = {
    'html'  : _EB_html,
    'latex' : _EB_latex,
    'text'  : _EB_text,
    'ppt'   : _EB_ppt }

_VEB_latex = _EBFormatter(_PrecisionFormatter(latex), '%s $\pm$ %s')

FormatSet.formatDict['VecErrorBars'] = {
    'html'  : _EB_html,
    'latex' : _VEB_latex,
    'text'  : _EB_text,
    'ppt'   : _EB_ppt}

class _PiEBFormatter(_EBFormatter):
    def __call__(self, t):
        if str(t[0]) == '--' or str(t[0]) == '':  return t[0]
        else:
            return super(_PiEBFormatter, self).__call__(t)

_PiEB_latex = _PiEBFormatter(_PrecisionFormatter(latex),
                           '$ \\begin{array}{c}(%s \\\\ \\pm %s)\\pi \\end{array} $',
                           '%s$\\pi$')
def _pi_eb_template(f):
    return _EBFormatter(_PrecisionFormatter(html), '(%s +/- %s)&pi')

# 'errorbars with pi' formatting: display (scalar_value +/- error bar) * pi
FormatSet.formatDict['PiErrorBars'] = {
    'html'  : _pi_eb_template(html),
    'latex' : _PiEB_latex,
    'text'  : _EB_text,
    'ppt'   : _pi_eb_template(ppt)}

FormatSet.formatDict['GateString'] = {
    'html'  : lambda s : '.'.join(s) if s is not None else '',
    'latex' : lambda s : ''          if s is None else ('$%s$' % '\\cdot'.join([ ('\\mbox{%s}' % gl) for gl in s])),
    'text'  : lambda s : tuple(s)    if s is not None else '',
    'ppt'   : lambda s : '.'.join(s) if s is not None else ''}

# 'pre' formatting, where the user gives the data in separate formats
def _pre_fmt_template(formatname):
    return lambda label : label[formatname]

FormatSet.formatDict['Pre'] = {
    'html'   : _pre_fmt_template('html'),
    'latex'  : _pre_fmt_template('latex'),
    'text'   : _pre_fmt_template('text'),
    'ppt'    : _pre_fmt_template('ppt')}


FormatSet.formatDict['Figure'] = {
    'html'  : _FigureFormatter(formatstring="<img width='%.2f' height='%.2f' src='%s/%s'>",
                               extension='.png'),
    'latex' : _FigureFormatter(formatstring="\\vcenteredhbox{\\includegraphics[width=%.2fin,height=%.2fin,keepaspectratio]{%s/%s}}",
                               extension='.pdf'),
    'text'  : lambda figInfo : figInfo[0],
    'ppt'   : lambda figInfo : 'Figure formatting not implemented for ppt'} # Not Implemented

# Bold formatting
FormatSet.formatDict['Bold'] = {
    'html'  : _PrecisionFormatter(html, formatstring='<b>%s</b>'),
    'latex' : _PrecisionFormatter(latex, formatstring='\\textbf{%s}'),
    'text'  : _Formatter(formatstring='**%s**'),
    'ppt'   : _PrecisionFormatter(ppt)} # No bold in ppt?


#Multi-row and multi-column formatting (with "Conversion" type inner formatting)
FormatSet.formatDict['MultiRow'] = {
    'html'  : _TupleFormatter(_fmtCnv_html),
    'latex' : _TupleFormatter(_fmtCnv_latex, formatstring='\\multirow{{{l1}}}{{*}}{{{l0}}}'),
    'text'  : _TupleFormatter(_Formatter(stringreplacers=[('<STAR>', '*'), ('|', ' ')])),
    'ppt'   : _TupleFormatter(_Formatter(stringreplacers=[('<STAR>', '*'), ('|', '\n')]))}


def _empty_str(l): return ""
FormatSet.formatDict['SpannedRow'] = {
    'html'  : _fmtCnv_html,
    'latex' : _empty_str,
    'text'  : _Formatter(stringreplacers=[('<STAR>', '*'), ('|', ' ')]),
    'ppt'   : _Formatter(stringreplacers=[('<STAR>', '*'), ('|', '\n')])}

def _repeat_no_format(label_tuple): 
    label, reps = label_tuple
    return ["%s" % label]*reps

FormatSet.formatDict['MultiCol'] = {
    'html'  : _repeat_no_format,
    'latex' : _TupleFormatter(_fmtCnv_latex, formatstring='\\multicolumn{{{l1}}}{{c|}}{{{l0}}}'),
    'text'  : _repeat_no_format,
    'ppt'   : _repeat_no_format}


#Special formatting for Hamiltonian and Stochastic gateset types
FormatSet.formatDict['GatesetType'] = {
    'html'  : _no_format,
    'latex' : _Formatter(stringreplacers=[('H','$\\mathcal{H}$'),('S','$\\mathcal{S}$')]),
    'text'  : _no_format,
    'ppt'   : _no_format}



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Functions for loading GST objects from text files."""

import os as _os
import json as _json

from . import stdinput as _stdinput
from .. import objects as _objs

def load_parameter_file(filename):
    """
    Load a json-formatted parameter file.

    Parameters
    ----------
    filename : string
        The name of the file to load.

    Returns
    -------
    dict
        The json file converted to a python dictionary.
    """
    with open(filename, 'r') as inputfile:
        return _json.load(inputfile)
    # return _json.load( open(filename, "rb") )

def load_dataset(filename, cache=False, collisionAction="aggregate",
                 verbosity=1):
    """
    Load a DataSet from a file.  First tries to load file as a
    saved DataSet object, then as a standard text-formatted DataSet.

    Parameters
    ----------
    filename : string
        The name of the file

    cache : bool, optional
        When set to True, a pickle file with the name filename + ".cache"
        is searched for and loaded instead of filename if it exists
        and is newer than filename.  If no cache file exists or one
        exists but it is older than filename, a cache file will be
        written after loading from filename.

    collisionAction : {"aggregate", "keepseparate"}
        Specifies how duplicate gate sequences should be handled.  "aggregate"
        adds duplicate-sequence counts, whereas "keepseparate" tags duplicate-
        sequence data with by appending a final "#<number>" gate label to the
        duplicated gate sequence.

    verbosity : int, optional
        If zero, no output is shown.  If greater than zero,
        loading progress is shown.

    Returns
    -------
    DataSet
    """

    printer = _objs.VerbosityPrinter.build_printer(verbosity)
    try:
        # a saved Dataset object is ok
        ds = _objs.DataSet(fileToLoadFrom=filename)
    except:

        #Parser functions don't take a VerbosityPrinter yet, and so
        # always output to stdout (TODO)
        bToStdout = (printer.verbosity > 0 and printer.filename is None)

        if cache:
            #bReadCache = False
            cache_filename = filename + ".cache"
            if _os.path.exists( cache_filename ) and \
               _os.path.getmtime(filename) < _os.path.getmtime(cache_filename):
                try:
                    printer.log("Loading from cache file: %s" % cache_filename)
                    ds = _objs.DataSet(fileToLoadFrom=cache_filename)
                    return ds
                except: print("WARNING: Failed to load from cache file")
            else:
                printer.log("Cache file not found or is tool old -- one will"
                            + "be created after loading is completed")

            # otherwise must use standard dataset file format
            parser = _stdinput.StdInputParser()
            ds = parser.parse_datafile(filename, bToStdout,
                                       collisionAction=collisionAction)

            printer.log("Writing cache file (to speed future loads): %s"
                        % cache_filename)
            ds.save(cache_filename)
        else:
            # otherwise must use standard dataset file format
            parser = _stdinput.StdInputParser()
            ds = parser.parse_datafile(filename, bToStdout,
                                       collisionAction=collisionAction)
        return ds


def load_multidataset(filename, cache=False, collisionAction="aggregate",
                      verbosity=1):
    """
    Load a MultiDataSet from a file.  First tries to load file as a
    saved MultiDataSet object, then as a standard text-formatted MultiDataSet.

    Parameters
    ----------
    filename : string
        The name of the file

    cache : bool, optional
        When set to True, a pickle file with the name filename + ".cache"
        is searched for and loaded instead of filename if it exists
        and is newer than filename.  If no cache file exists or one
        exists but it is older than filename, a cache file will be
        written after loading from filename.

    collisionAction : {"aggregate", "keepseparate"}
        Specifies how duplicate gate sequences should be handled.  "aggregate"
        adds duplicate-sequence counts, whereas "keepseparate" tags duplicate-
        sequence data with by appending a final "#<number>" gate label to the
        duplicated gate sequence.

    verbosity : int, optional
        If zero, no output is shown.  If greater than zero,
        loading progress is shown.


    Returns
    -------
    MultiDataSet
    """

    printer = _objs.VerbosityPrinter.build_printer(verbosity)
    try:
        # a saved MultiDataset object is ok
        mds = _objs.MultiDataSet(fileToLoadFrom=filename)
    except:

        #Parser functions don't take a VerbosityPrinter yet, and so
        # always output to stdout (TODO)
        bToStdout = (printer.verbosity > 0 and printer.filename is None)

        if cache:
            # bReadCache = False
            cache_filename = filename + ".cache"
            if _os.path.exists( cache_filename ) and \
               _os.path.getmtime(filename) < _os.path.getmtime(cache_filename):
                try:
                    printer.log("Loading from cache file: %s" % cache_filename)
                    mds = _objs.MultiDataSet(fileToLoadFrom=cache_filename)
                    return mds
                except: print("WARNING: Failed to load from cache file")
            else:
                printer.log("Cache file not found or is too old -- one will be"
                            + "created after loading is completed")

            # otherwise must use standard dataset file format
            parser = _stdinput.StdInputParser()
            mds = parser.parse_multidatafile(filename, bToStdout,
                                             collisionAction=collisionAction)

            printer.log("Writing cache file (to speed future loads): %s" 
                        % cache_filename)
            mds.save(cache_filename)

        else:
            # otherwise must use standard dataset file format
            parser = _stdinput.StdInputParser()
            mds = parser.parse_multidatafile(filename, bToStdout,
                                             collisionAction=collisionAction)
    return mds



def load_gateset(filename):
    """
    Load a GateSet from a file, formatted using the
    standard text-format for gate sets.

    Parameters
    ----------
    filename : string
        The name of the file

    Returns
    -------
    GateSet
    """
    return _stdinput.read_gateset(filename)

def load_gatestring_dict(filename):
    """
    Load a gate string dictionary from a file, formatted
    using the standard text-format.

    Parameters
    ----------
    filename : string
        The name of the file.

    Returns
    -------
    Dictionary with keys = gate string labels and
      values = GateString objects.
    """
    std = _stdinput.StdInputParser()
    return std.parse_dictfile(filename)

def load_gatestring_list(filename, readRawStrings=False):
    """
    Load a gate string list from a file, formatted
    using the standard text-format.

    Parameters
    ----------
    filename : string
        The name of the file

    readRawStrings : boolean
        If True, gate strings are not converted
        to tuples of gate labels.

    Returns
    -------
    list of GateString objects
    """
    if readRawStrings:
        rawList = []
        with open(filename, 'r') as gatestringlist:
            for line in gatestringlist:
                if len(line.strip()) == 0: continue
                if len(line) == 0 or line[0] == '#': continue
                rawList.append( line.strip() )
        return rawList
    else:
        std = _stdinput.StdInputParser()
        return std.parse_stringfile(filename)
